import cv2

# http://planning.cs.uiuc.edu/node102.html
def get_rotate_matrix(x, y, z):
    """
    按照 zyx 的顺序旋转，输入角度单位为 degrees, 均为顺时针旋转
    :param x: X-axis
    :param y: Y-axis
    :param z: Z-axis
    :return:
    """
    x = math.radians(x)
    y = math.radians(y)
    z = math.radians(z)

    c, s = math.cos(y), math.sin(y)
    M_y = np.matrix([[c, 0., s, 0.],
                     [0., 1., 0., 0.],
                     [-s, 0., c, 0.],
                     [0., 0., 0., 1.]])

    c, s = math.cos(x), math.sin(x)
    M_x = np.matrix([[1., 0., 0., 0.],
                     [0., c, -s, 0.],
                     [0., s, c, 0.],
                     [0., 0., 0., 1.]])

    c, s = math.cos(z), math.sin(z)
    M_z = np.matrix([[c, -s, 0., 0.],
                     [s, c, 0., 0.],
                     [0., 0., 1., 0.],
                     [0., 0., 0., 1.]])

    return M_x * M_y * M_z


def cliped_rand_norm(mu=0, sigma3=1):
    """
    :param mu: 均值
    :param sigma3: 3 倍标准差， 99% 的数据落在 (mu-3*sigma, mu+3*sigma)
    :return:
    """
    # 标准差
    sigma = sigma3 / 3
    dst = sigma * np.random.randn() + mu
    dst = np.clip(dst, 0 - sigma3, sigma3)
    return dst


def warpPerspective(src, M33, sl, gpu):
    if gpu:
        from libs.gpu.GpuWrapper import cudaWarpPerspectiveWrapper
        dst = cudaWarpPerspectiveWrapper(src.astype(np.uint8), M33, (sl, sl), cv2.INTER_CUBIC)
    else:
        dst = cv2.warpPerspective(src, M33, (sl, sl), flags=cv2.INTER_CUBIC)
    return dst


# https://stackoverflow.com/questions/17087446/how-to-calculate-perspective-transform-for-opencv-from-rotation-angles
# https://nbviewer.jupyter.org/github/manisoftwartist/perspectiveproj/blob/master/perspective.ipynb
# http://planning.cs.uiuc.edu/node102.html
class PerspectiveTransform(object):
    def __init__(self, x, y, z, scale, fovy):
        self.x = x
        self.y = y
        self.z = z
        self.scale = scale
        self.fovy = fovy

    def transform_image(self, src, gpu=False):
        if len(src.shape) > 2:
            H, W, C = src.shape
        else:
            H, W = src.shape

        M33, sl, _, ptsOut = self.get_warp_matrix(W, H, self.x, self.y, self.z, self.scale, self.fovy)
        sl = int(sl)

        dst = warpPerspective(src, M33, sl, gpu)

        return dst, M33, ptsOut

    def transform_pnts(self, pnts, M33):
        """
        :param pnts: 2D pnts, left-top, right-top, right-bottom, left-bottom
        :param M33: output from transform_image()
        :return: 2D pnts apply perspective transform
        """
        pnts = np.asarray(pnts, dtype=np.float32)
        pnts = np.array([pnts])
        dst_pnts = cv2.perspectiveTransform(pnts, M33)[0]

        return dst_pnts

    def get_warped_pnts(self, ptsIn, ptsOut, W, H, sidelength):
        ptsIn2D = ptsIn[0, :]
        ptsOut2D = ptsOut[0, :]
        ptsOut2Dlist = []
        ptsIn2Dlist = []

        for i in range(0, 4):
            ptsOut2Dlist.append([ptsOut2D[i, 0], ptsOut2D[i, 1]])
            ptsIn2Dlist.append([ptsIn2D[i, 0], ptsIn2D[i, 1]])

        pin = np.array(ptsIn2Dlist) + [W / 2., H / 2.]
        pout = (np.array(ptsOut2Dlist) + [1., 1.]) * (0.5 * sidelength)
        pin = pin.astype(np.float32)
        pout = pout.astype(np.float32)

        return pin, pout

    def get_warp_matrix(self, W, H, x, y, z, scale, fV):
        fVhalf = np.deg2rad(fV / 2.)
        d = np.sqrt(W * W + H * H)
        sideLength = scale * d / np.cos(fVhalf)
        h = d / (2.0 * np.sin(fVhalf))
        n = h - (d / 2.0)
        f = h + (d / 2.0)

        # Translation along Z-axis by -h
        T = np.eye(4, 4)
        T[2, 3] = -h

        # Rotation matrices around x,y,z
        R = get_rotate_matrix(x, y, z)

        # Projection Matrix
        P = np.eye(4, 4)
        P[0, 0] = 1.0 / np.tan(fVhalf)
        P[1, 1] = P[0, 0]
        P[2, 2] = -(f + n) / (f - n)
        P[2, 3] = -(2.0 * f * n) / (f - n)
        P[3, 2] = -1.0

        # pythonic matrix multiplication
        M44 = reduce(lambda x, y: np.matmul(x, y), [P, T, R])

        # shape should be 1,4,3 for ptsIn and ptsOut since perspectiveTransform() expects data in this way.
        # In C++, this can be achieved by Mat ptsIn(1,4,CV_64FC3);
        ptsIn = np.array([[
            [-W / 2., H / 2., 0.],
            [W / 2., H / 2., 0.],
            [W / 2., -H / 2., 0.],
            [-W / 2., -H / 2., 0.]
        ]])
        ptsOut = cv2.perspectiveTransform(ptsIn, M44)

        ptsInPt2f, ptsOutPt2f = self.get_warped_pnts(ptsIn, ptsOut, W, H, sideLength)

        # check float32 otherwise OpenCV throws an error
        assert (ptsInPt2f.dtype == np.float32)
        assert (ptsOutPt2f.dtype == np.float32)
        M33 = cv2.getPerspectiveTransform(ptsInPt2f, ptsOutPt2f).astype(np.float32)

        return M33, sideLength, ptsInPt2f, ptsOutPt2f

def apply_perspective_transform(img, max_x, max_y, max_z):
    x = cliped_rand_norm(0, max_x)
    y = cliped_rand_norm(0, max_y)
    z = cliped_rand_norm(0, max_z)

    transformer = PerspectiveTransform(x, y, z, scale=1.0, fovy=50)
    dst_img, M33, dst_img_pnts = transformer.transform_image(img, False)
    dst_img_pnts = np.array(dst_img_pnts,dtype = np.int32)

    x_min, y_min = np.min(dst_img_pnts,axis=0)
    x_max, y_max = np.max(dst_img_pnts,axis=0)
    return dst_img[y_min:y_max, x_min:x_max]


import os
import re
import random
from collections import OrderedDict
from collections import defaultdict
import cv2
from ..utils import LOG
from .factory import AugFactory

class Chain(object):
    def __init__(self, flow_list):
        self.func_dict = dict()
        self.initChainKind()
        self.auditChainKind(flow_list)
        flow = flow_list.split(self.chain_kind)
        self.op_sym_list = [x.strip() for x in flow]
        self.p_list = [ float(x.split('@')[1].strip()) if '@' in x else 1 for x in self.op_sym_list  ]
        self.op_sym_list = [x.split('@')[0].strip() for x in self.op_sym_list if x]
        self.op_list = []

    def __call__(self, img):
        return self.func_dict[self.chain_kind](img)

    def addOp(self, op, p=1):
        self.op_list.append(op)
        self.p_list.append(p)

    def initChainKind(self):
        self.func_dict['=>'] = self.process

    def process(self, img):
        for t in self.op_list:
            img = t(img)
        return img

    def auditChainKind(self,flow_list):
        self.chain_kind = '=>'
        tokens = re.sub('[a-zA-Z0-9.@]'," ", flow_list).split()

        if len(tokens) == 0:
            return

        tokens = set(tokens)

        if len(tokens) > 1:
            raise Exception('Multi token found in flow list: ', tokens)

        self.chain_kind = tokens.pop()

        if self.chain_kind not in self.func_dict.keys():
            raise Exception("token not supported: ", self.chain_kind)

class AugChain(Chain):
    def __init__(self, flow_list, deepvac_config):
        super(AugChain, self).__init__(flow_list)
        self.factory = AugFactory()
        self.op_list = [self.factory.get(x)(deepvac_config) for x in self.op_sym_list]

    def initChainKind(self):
        self.func_dict['=>'] = self.processRandom
        self.func_dict['||'] = self.processDice

    def processRandom(self, img):
        for t,p in zip(self.op_list, self.p_list):
            if random.random() < p:
                img = t(img)
        return img

    def processDice(self, img):
        i = random.randrange(len(self.op_list))
        return self.op_list[i](img)

class Composer(object):
    def __init__(self,deepvac_config):
        self._graph = OrderedDict()
        self._graph_p = OrderedDict()
        self.config = deepvac_config
        self.auditConfig()

    def addAugChain(self, name, chain, p=1.0):
        self._graph[name] = chain
        self._graph_p[name] = p

    def auditConfig(self):
        pass

    def addOp(self, name, op, p=1.0):
        self._graph[name].addOp(op,p)

    def remove(self, name):
        try:
            del self._graph[name]
        except KeyError:
            pass

    def __call__(self, img):
        for k in self._graph:
            if random.random() < self._graph_p[k]:
                img = self._graph[k](img)
        return img

class DiceComposer(Composer):
    def __call__(self, img):
        i = random.randrange(len(self._graph_p))
        return list(self._graph.values())[i]

class PickOneComposer(Composer):
    def __call__(self, img):
        for k in self._graph:
            if random.random() >= self._graph_p[k]:
                continue
            return self._graph[k](img)
        return self._graph[k](img)

class FaceAugComposer(Composer):
    def __init__(self, deepvac_config):
        super(FaceAugComposer, self).__init__(deepvac_config)
        ac1 = AugChain('RandomColorJitterAug@0.5 => MosaicAug@0.5',deepvac_config)
        ac2 = AugChain('IPCFaceAug || MotionAug',deepvac_config)
        self.addAugChain('ac1', ac1, 1)
        self.addAugChain('ac2', ac2, 0.5)

class OcrAugComposer(Composer):
    def __init__(self, deepvac_config):
        super(OcrAugComposer, self).__init__(deepvac_config)

        ac1 = AugChain('SpeckleAug@0.2 => HorlineAug@0.2 => NoisyAug@0.2',deepvac_config)
        ac2 = AugChain('MotionAug || AffineAug || PerspectAug || GaussianAug || VerlineAug || LRmotionAug || UDmotionAug || PerspectiveAug || StretchAug',deepvac_config)

        self.addAugChain('ac1', ac1, 0.2)
        self.addAugChain('ac2', ac2, 0.9)

class YoloAugComposer(Composer):
    def __init__(self, deepvac_config):
        super(YoloAugComposer, self).__init__(deepvac_config)
        if deepvac_config.hflip is None:
            deepvac_config.hflip = 0.5
        ac = AugChain("YoloPerspectiveAug => HSVAug => YoloNormalizeAug => YoloHFlipAug@{}".format(deepvac_config.hflip), deepvac_config)
        self.addAugChain("ac", ac)

class RetinaAugComposer(PickOneComposer):
    def __init__(self, deepvac_config):
        super(RetinaAugComposer, self).__init__(deepvac_config)
        ac1 = AugChain('CropFacialWithBoxesAndLmksAug => BrightDistortFacialAug@0.5 => ContrastDistortFacialAug@0.5 => SaturationDistortFacialAug@0.5 \
                => HueDistortFacialAug@0.5 => Pad2SquareFacialAug => MirrorFacialAug@0.5 => ResizeSubtractMeanFacialAug', deepvac_config)
        ac2 = AugChain('CropFacialWithBoxesAndLmksAug => BrightDistortFacialAug@0.5 => SaturationDistortFacialAug@0.5 => HueDistortFacialAug@0.5 \
        => ContrastDistortFacialAug@0.5 => Pad2SquareFacialAug => MirrorFacialAug@0.5 => ResizeSubtractMeanFacialAug', deepvac_config)
        self.addAugChain("ac1", ac1, 0.5)
        self.addAugChain("ac2", ac2, 0.5)

class OcrDetectAugComposer(Composer):
    def __init__(self, deepvac_config):
        super(OcrDetectAugComposer, self).__init__(deepvac_config)
        ac1 = AugChain('ImageWithMasksRandomHorizontalFlipAug@0.5 => ImageWithMasksRandomRotateAug => ImageWithMasksRandomCropAug',deepvac_config)
        self.addAugChain('ac1', ac1, 1)

class SegImageAugComposer(Composer):
    def __init__(self, deepvac_config):
        super(SegImageAugComposer, self).__init__(deepvac_config)
        ac1 = AugChain('GaussianAug || RandomColorJitterAug || BrightnessJitterAug || ContrastJitterAug',deepvac_config)
        ac2 = AugChain('MotionAug',deepvac_config)
        self.addAugChain('ac1', ac1, 1)
        self.addAugChain('ac2', ac1, 0.2)

class SegImageWithMaskAugComposer(Composer):
    def __init__(self, deepvac_config):
        super(SegImageWithMaskAugComposer, self).__init__(deepvac_config)
        ac1 = AugChain('ImageWithMasksRandomCropAug',deepvac_config)
        ac2 = AugChain('ImageWithMasksRandomHorizontalFlipAug || ImageWithMasksRandomRotateAug',deepvac_config)
        self.addAugChain('ac1', ac1, 0.4)
        self.addAugChain('ac2', ac2, 0.6)

if __name__ == "__main__":
    x = Chain("RandomColorJitterAug@0.3 => MosaicAug@0.8 => MotionAug ")
    LOG.logI(x.op_sym_list)
    LOG.logI(x.p_list)
    LOG.logI(x.chain_kind)



from ..utils import LOG

def makeDivisible(v, divisor=8, min_value=None):
    if min_value is None:
        min_value = divisor
    new_v = max(min_value, int(v + divisor / 2) // divisor * divisor)
    # Make sure that round down does not go down by more than 10%.
    if new_v < 0.9 * v:
        new_v += divisor
    return new_v

def separateBN4OptimizerPG(modules):
    paras_only_bn = []
    paras_wo_bn = []
    memo = set()
    gemfield_set = set()
    gemfield_set.update(set(modules.parameters()))
    LOG.logI("separateBN4OptimizerPG set len: {}".format(len(gemfield_set)))
    named_modules = modules.named_modules(prefix='')
    for module_prefix, module in named_modules:
        if "module" not in module_prefix:
            LOG.logI("separateBN4OptimizerPG skip {}".format(module_prefix))
            continue

        members = module._parameters.items()
        for k, v in members:
            name = module_prefix + ('.' if module_prefix else '') + k
            if v is None:
                continue
            if v in memo:
                continue
            memo.add(v)
            if "batchnorm" in str(module.__class__):
                paras_only_bn.append(v)
            else:
                paras_wo_bn.append(v)

    LOG.logI("separateBN4OptimizerPG param len: {} - {}".format(len(paras_wo_bn),len(paras_only_bn)))
    return paras_only_bn, paras_wo_bn







import torch
import numpy as np
import torch
from itertools import product as product
import numpy as np
from math import ceil


# Adapted from https://github.com/Hakuyume/chainer-ssd
def decode(loc, priors, variances):
    """Decode locations from predictions using priors to undo
    the encoding we did for offset regression at train time.
    Args:
        loc (tensor): location predictions for loc layers,
            Shape: [num_priors,4]
        priors (tensor): Prior boxes in center-offset form.
            Shape: [num_priors,4].
        variances: (list[float]) Variances of priorboxes
    Return:
        decoded bounding box predictions
    """

    boxes = torch.cat((
        priors[:, :2] + loc[:, :2] * variances[0] * priors[:, 2:],
        priors[:, 2:] * torch.exp(loc[:, 2:] * variances[1])), 1)
    boxes[:, :2] -= boxes[:, 2:] / 2
    boxes[:, 2:] += boxes[:, :2]
    return boxes

def decode_landm(pre, priors, variances):
    """Decode landm from predictions using priors to undo
    the encoding we did for offset regression at train time.
    Args:
        pre (tensor): landm predictions for loc layers,
            Shape: [num_priors,10]
        priors (tensor): Prior boxes in center-offset form.
            Shape: [num_priors,4].
        variances: (list[float]) Variances of priorboxes
    Return:
        decoded landm predictions
    """
    landms = torch.cat((priors[:, :2] + pre[:, :2] * variances[0] * priors[:, 2:],
                        priors[:, :2] + pre[:, 2:4] * variances[0] * priors[:, 2:],
                        priors[:, :2] + pre[:, 4:6] * variances[0] * priors[:, 2:],
                        priors[:, :2] + pre[:, 6:8] * variances[0] * priors[:, 2:],
                        priors[:, :2] + pre[:, 8:10] * variances[0] * priors[:, 2:],
                        ), dim=1)
    return landms

# Original author: Francisco Massa:
# https://github.com/fmassa/object-detection.torch
# Ported to PyTorch by Max deGroot (02/01/2017)
def nms(boxes, scores, overlap=0.5, top_k=200):
    """Apply non-maximum suppression at test time to avoid detecting too many
    overlapping bounding boxes for a given object.
    Args:
        boxes: (tensor) The location preds for the img, Shape: [num_priors,4].
        scores: (tensor) The class predscores for the img, Shape:[num_priors].
        overlap: (float) The overlap thresh for suppressing unnecessary boxes.
        top_k: (int) The Maximum number of box preds to consider.
    Return:
        The indices of the kept boxes with respect to num_priors.
    """

    keep = torch.Tensor(scores.size(0)).fill_(0).long()
    if boxes.numel() == 0:
        return keep
    x1 = boxes[:, 0]
    y1 = boxes[:, 1]
    x2 = boxes[:, 2]
    y2 = boxes[:, 3]
    area = torch.mul(x2 - x1, y2 - y1)
    v, idx = scores.sort(0)  # sort in ascending order
    # I = I[v >= 0.01]
    idx = idx[-top_k:]  # indices of the top-k largest vals
    xx1 = boxes.new()
    yy1 = boxes.new()
    xx2 = boxes.new()
    yy2 = boxes.new()
    w = boxes.new()
    h = boxes.new()

    # keep = torch.Tensor()
    count = 0
    while idx.numel() > 0:
        i = idx[-1]  # index of current largest val
        # keep.append(i)
        keep[count] = i
        count += 1
        if idx.size(0) == 1:
            break
        idx = idx[:-1]  # remove kept element from view
        # load bboxes of next highest vals
        torch.index_select(x1, 0, idx, out=xx1)
        torch.index_select(y1, 0, idx, out=yy1)
        torch.index_select(x2, 0, idx, out=xx2)
        torch.index_select(y2, 0, idx, out=yy2)
        # store element-wise max with next highest score
        xx1 = torch.clamp(xx1, min=x1[i])
        yy1 = torch.clamp(yy1, min=y1[i])
        xx2 = torch.clamp(xx2, max=x2[i])
        yy2 = torch.clamp(yy2, max=y2[i])
        w.resize_as_(xx2)
        h.resize_as_(yy2)
        w = xx2 - xx1
        h = yy2 - yy1
        # check sizes of xx1 and xx2.. after each iteration
        w = torch.clamp(w, min=0.0)
        h = torch.clamp(h, min=0.0)
        inter = w*h
        # IoU = i / (area(a) + area(b) - i)
        rem_areas = torch.index_select(area, 0, idx)  # load remaining areas)
        union = (rem_areas - inter) + area[i]
        IoU = inter/union  # store result in iou
        # keep only elements with an IoU <= overlap
        idx = idx[IoU.le(overlap)]
    return keep, count

def py_cpu_nms(dets, thresh):
    """Pure Python NMS baseline."""
    x1 = dets[:, 0]
    y1 = dets[:, 1]
    x2 = dets[:, 2]
    y2 = dets[:, 3]
    scores = dets[:, 4]

    areas = (x2 - x1 + 1) * (y2 - y1 + 1)
    order = scores.argsort()[::-1]

    keep = []
    while order.size > 0:
        i = order[0]
        keep.append(i)
        xx1 = np.maximum(x1[i], x1[order[1:]])
        yy1 = np.maximum(y1[i], y1[order[1:]])
        xx2 = np.minimum(x2[i], x2[order[1:]])
        yy2 = np.minimum(y2[i], y2[order[1:]])

        w = np.maximum(0.0, xx2 - xx1 + 1)
        h = np.maximum(0.0, yy2 - yy1 + 1)
        inter = w * h
        ovr = inter / (areas[i] + areas[order[1:]] - inter)

        inds = np.where(ovr <= thresh)[0]
        order = order[inds + 1]

    return keep

class PriorBox(object):
    def __init__(self, cfg, image_size=None, phase='train'):
        super(PriorBox, self).__init__()
        self.min_sizes = cfg['min_sizes']
        self.steps = cfg['steps']
        self.clip = cfg['clip']
        self.image_size = image_size
        self.feature_maps = [[ceil(self.image_size[0]/step), ceil(self.image_size[1]/step)] for step in self.steps]
        self.name = "s"

    def forward(self):
        anchors = []
        for k, f in enumerate(self.feature_maps):
            min_sizes = self.min_sizes[k]
            for i, j in product(range(f[0]), range(f[1])):
                for min_size in min_sizes:
                    s_kx = min_size / self.image_size[1]
                    s_ky = min_size / self.image_size[0]
                    dense_cx = [x * self.steps[k] / self.image_size[1] for x in [j + 0.5]]
                    dense_cy = [y * self.steps[k] / self.image_size[0] for y in [i + 0.5]]
                    for cy, cx in product(dense_cy, dense_cx):
                        anchors += [cx, cy, s_kx, s_ky]

        # back to torch land
        output = torch.Tensor(anchors).view(-1, 4)
        if self.clip:
            output.clamp_(max=1, min=0)
        return output



from .activate_layer import *
from .bottleneck_layer import *
from .conv_layer import *
from .fpn_layer import *
from .op_layer import *
from .res_layer import *
from .se_layer import *
from .spp_layer import *
from .ssh_layer import *
from .weights_init import *

from .mobilenet import *
from .regnet import *
from .repvgg import *
from .resnet import *



import numpy as np
import random
from .base_aug import AugBase
from .perspective_helper import apply_perspective_transform
from .remaper_helper import Remaper
from .line_helper import Liner
from .emboss_helper import apply_emboss

class TextRendererPerspectiveAug(AugBase):
    def __init__(self, deepvac_config):
        super(TextRendererPerspectiveAug, self).__init__(deepvac_config)

    def auditConfig(self):
        self.config.text_renderer_perspective_max_x = self.addUserConfig('text_renderer_perspective_max_x', self.config.text_renderer_perspective_max_x, 10)
        self.config.text_renderer_perspective_max_y = self.addUserConfig('text_renderer_perspective_max_y', self.config.text_renderer_perspective_max_y, 10)
        self.config.text_renderer_perspective_max_z = self.addUserConfig('text_renderer_perspective_max_z', self.config.text_renderer_perspective_max_z, 5)

    def __call__(self, img):
        self.auditInput(img)
        return apply_perspective_transform(img, self.config.text_renderer_perspective_max_x, self.config.text_renderer_perspective_max_y, self.config.text_renderer_perspective_max_z)

class TextRendererCurveAug(AugBase):
    def __init__(self, deepvac_config):
        super(TextRendererCurveAug, self).__init__(deepvac_config)

    def auditConfig(self):
        pass

    def __call__(self, img):
        self.auditInput(img)
        h, w = img.shape[:2]
        re_img, text_box_pnts = Remaper().apply(img, [[0,0],[w,0],[w,h],[0,h]])
        return re_img

class TextRendererLineAug(AugBase):
    def __init__(self, deepvac_config):
        super(TextRendererLineAug, self).__init__(deepvac_config)

    def auditConfig(self):
        self.config.text_renderer_line_offset = self.addUserConfig('text_renderer_line_offset', self.config.text_renderer_line_offset, 5)

    def __call__(self, img):
        self.auditInput(img)
        h, w = img.shape[:2]
        offset = self.config.text_renderer_line_offset
        pos = [[offset,offset],[w-offset,offset],[w-offset,h-offset],[offset,h-offset]]
        re_img, text_box_pnts = Liner().apply(img, pos)
        return re_img

class TextRendererEmbossAug(AugBase):
    def __init__(self, deepvac_config):
        super(TextRendererEmbossAug, self).__init__(deepvac_config)

    def auditConfig(self):
        pass

    def __call__(self, img):
        self.auditInput(img)
        return apply_emboss(img)

class TextRendererReverseAug(AugBase):
    def __init__(self, deepvac_config):
        super(TextRendererReverseAug, self).__init__(deepvac_config)

    def auditConfig(self):
        pass

    def __call__(self, img):
        self.auditInput(img)
        offset = np.random.randint(-10, 10)
        return 255 + offset - img


import torch.nn as nn
from .module_helper import makeDivisible
from .activate_layer import hsigmoid

class SELayer(nn.Module):
    def __init__(self, channel, reduction=4):
        super(SELayer, self).__init__()
        self.avg_pool = nn.AdaptiveAvgPool2d(1)
        self.fc = nn.Sequential(
            nn.Linear(channel, makeDivisible(channel // reduction, 8)),
            nn.ReLU(inplace=True),
            nn.Linear(makeDivisible(channel // reduction, 8), channel),
            hsigmoid(inplace=True)
        )

    def forward(self, x):
        b, c, _, _ = x.size()
        y = self.avg_pool(x).view(b, c)
        y = self.fc(y).view(b, c, 1, 1)
        return x * y


import torch
import torch.nn as nn
from .conv_layer import Conv2dBNReLU

class BottleneckStd(nn.Module):
    # Standard bottleneck
    def __init__(self, in_planes, out_planes, groups=1, shortcut=True, eps=1e-5, momentum=0.1, expansion=0.5):  # ch_in, ch_out, shortcut, groups, expansion
        super(BottleneckStd, self).__init__()
        hidden_planes = int(out_planes * expansion)  # hidden channels
        self.conv1 = Conv2dBNHardswish(in_planes, hidden_planes, 1, 1, eps, momentum)
        self.conv2 = Conv2dBNHardswish(hidden_planes, out_planes, 3, 1, eps, momentum, groups=groups)
        self.add = shortcut and in_planes == out_planes

    def forward(self, x):
        return x + self.conv2(self.conv1(x)) if self.add else self.conv2(self.conv1(x))

class BottleneckCSP(nn.Module):
    def __init__(self, in_planes, out_planes, bottle_std_num=1, shortcut=True, eps=1e-5, momentum=0.1, groups=1, expansion=0.5):  # ch_in, ch_out, number, shortcut, groups, expansion
        super(BottleneckCSP, self).__init__()
        hidden_planes = int(out_planes * expansion)  # hidden channels
        self.conv1 = Conv2dBNHardswish(in_planes, hidden_planes, 1, 1, eps, momentum)
        self.conv2 = nn.Conv2d(in_planes, hidden_planes, 1, 1, bias=False)
        self.conv3 = nn.Conv2d(hidden_planes, hidden_planes, 1, 1, bias=False)
        self.conv4 = Conv2dBNHardswish(2 * hidden_planes, out_planes, 1, 1, eps, momentum)
        self.bn = nn.BatchNorm2d(2 * hidden_planes, eps=eps, momentum=momentum)  # applied to cat(conv2, conv3)
        self.act = nn.LeakyReLU(0.1, inplace=True)
        self.std_bottleneck_list = nn.Sequential(*[BottleneckStd(hidden_planes, hidden_planes, groups=groups, shortcut=shortcut, eps=eps, momentum=momentum, expansion=1.0) for _ in range(bottle_std_num)])

    def forward(self, x):
        y1 = self.conv3(self.std_bottleneck_list(self.conv1(x)))
        y2 = self.conv2(x)
        return self.conv4(self.act(self.bn(torch.cat((y1, y2), dim=1))))

class Bottleneck(nn.Module):
    expansion: int = 4
    def __init__(self, inplanes: int, outplanes: int, stride: int = 1):
        super(Bottleneck, self).__init__()
        self.conv1 = Conv2dBNReLU(in_planes=inplanes, out_planes=outplanes, kernel_size=1)
        self.conv2 = Conv2dBNReLU(in_planes=outplanes, out_planes=outplanes, kernel_size=3, stride=stride)

        outplanes_after_expansion = outplanes * self.expansion
        self.conv3 = nn.Conv2d(outplanes, outplanes_after_expansion, kernel_size=1)
        self.bn3 = nn.BatchNorm2d(outplanes_after_expansion)
        self.relu = nn.ReLU(inplace=True)

        self.stride = stride
        self.downsample = None

        if stride != 1 or inplanes != outplanes_after_expansion:
            self.downsample = nn.Sequential(nn.Conv2d(inplanes, outplanes_after_expansion, kernel_size=1, stride=stride, bias=False), nn.BatchNorm2d(outplanes_after_expansion))

    def forward(self, x):
        identity = x
        out = self.conv1(x)
        out = self.conv2(out)
        out = self.conv3(out)
        out = self.bn3(out)

        if self.downsample is not None:
            identity = self.downsample(x)

        out += identity
        return self.relu(out)

class BottleneckIR(nn.Module):
    def __init__(self, inplanes: int, outplanes: int, stride: int):
        super(BottleneckIR, self).__init__()

        self.shortcut_layer = nn.MaxPool2d(kernel_size=1, stride=stride) if inplanes == outplanes else nn.Sequential(
                nn.Conv2d(inplanes, outplanes, kernel_size=1, stride=stride ,bias=False), nn.BatchNorm2d(outplanes))

        self.res_layer = nn.Sequential(
            nn.BatchNorm2d(inplanes),
            nn.Conv2d(inplanes, outplanes, kernel_size=3, stride=1, padding=1 ,bias=False),
            nn.PReLU(outplanes),
            nn.Conv2d(outplanes, outplanes, kernel_size=3, stride=stride, padding=1 ,bias=False),
            nn.BatchNorm2d(outplanes)
        )

    def forward(self, x):
        shortcut = self.shortcut_layer(x)
        res = self.res_layer(x)
        return res + shortcut


import os
import argparse
import collections
import time
import copy
import math
from typing import Callable
import torch
import torch.nn as nn
import torch.optim as optim
import torch.distributed as dist
from torch.utils.data import DataLoader
from torch.cuda.amp import autocast
from torch.cuda.amp import GradScaler

try:
    from torch.utils.tensorboard import SummaryWriter
except ImportError:
    LOG.logE("Deepvac has dependency on tensorboard, please install tensorboard first, e.g. [pip3 install tensorboard]", exit=True)

from ..utils import syszux_once, LOG, assertAndGetGitBranch, getPrintTime, AverageMeter
from ..cast import export3rd

#deepvac implemented based on PyTorch Framework
class Deepvac(object):
    def __init__(self, deepvac_train_config):
        self.config = deepvac_train_config
        if self.config.is_forward_only is None:
            self.config.is_forward_only = True
        self.config.branch = assertAndGetGitBranch(deepvac_train_config.is_disable_git)
        self.init()

    def auditConfig(self):
        if not self.config.model_path and not self.config.jit_model_path:
            LOG.logE("both config.model_path and config.jit_model_path are not configured, cannot do predict.", exit=True)
        #audit for ema
        if self.config.is_forward_only and self.config.ema:
            LOG.logE("Error: You must disable config.ema in test only mode.", exit=True)

    def _parametersInfo(self):
        param_info_list = [p.numel() for p in self.config.net.parameters() ]
        LOG.logI("config.net has {} parameters.".format(sum(param_info_list)))
        LOG.logI("config.net parameters detail: {}".format(['{}: {}'.format(name, p.numel()) for name, p in self.config.net.named_parameters()]))

    def initDevice(self):
        #to determine CUDA device, different in DDP
        self.config.device = torch.device(self.config.device)

    def initNetWithCode(self):
        if self.config.net is None:
            LOG.logE("You must implement and configure config.train.net to a torch.nn.Module instance in config.py.", exit=True)
        if not isinstance(self.config.net, nn.Module):
            LOG.logE("You must configure config.train.net to a torch.nn.Module instance in config.py.", exit=True)

    def initStateDict(self):
        self.config.state_dict = None
        if not self.config.model_path:
            LOG.logI("config.model_path not specified, network parametes will not be initialized from trained/pretrained model.")
            return

        if self.config.jit_model_path and self.config.is_forward_only:
            LOG.logI("config.jit_model_path specified in forward-only mode, network parametes will be initialized from jit model rather than trained/pretrained model.")
            return

        if self.config.jit_model_path:
            LOG.logW("config.jit_model_path specified in training mode, omit...")

        LOG.logI('Loading State Dict from {}'.format(self.config.model_path))
        self.config.state_dict = torch.load(self.config.model_path, map_location=self.config.device)
        #remove prefix begin
        prefix = 'module.'
        f = lambda x: x.split(prefix, 1)[-1] if x.startswith(prefix) else x
        if "state_dict" in self.config.state_dict.keys():
            self.config.state_dict = {f(key): value for key, value in self.config.state_dict['state_dict'].items()}
        else:
            self.config.state_dict = {f(key): value for key, value in self.config.state_dict.items()}
        #remove prefix end

        #audit on model file
        state_dict_keys = set(self.config.state_dict.keys())
        code_net_keys = set(self.config.net.state_dict().keys())
        used_keys = code_net_keys & state_dict_keys
        unused_keys = state_dict_keys - code_net_keys
        missing_keys = code_net_keys - state_dict_keys
        LOG.logI('Missing keys:{} | {}'.format(len(missing_keys), missing_keys))
        LOG.logI('Unused keys:{} | {}'.format(len(unused_keys), unused_keys))
        LOG.logI('Used keys:{}'.format(len(used_keys)))

        if not self.config.model_reinterpret_cast and len(used_keys) == 0:
            LOG.logE('Error: load NONE from pretrained model: {}'.format(self.config.model_path), exit=True)

        if len(missing_keys) > 0:
            LOG.logW("There have missing network parameters, double check if you are using a mismatched trained model.")
            if not self.config.network_audit_disabled:
                LOG.logE("If you know this risk, set config.network_audit_disabled=True in config.py to omit this error.")

    def castStateDict(self):
        LOG.logI("config.model_reinterpret_cast is True, Try to reinterpret cast the model")
        state_dict = collections.OrderedDict()
        keys = list(self.config.state_dict.keys())
        for idx, name in enumerate(self.config.net.state_dict()):
            if self.config.net.state_dict()[name].size() != self.config.state_dict[keys[idx]].size():
                LOG.logE("Weights shape must be equal, {} and {} shape are not equal".format(keys[idx], name), exit=True)
            state_dict[name] = self.config.state_dict[keys[idx]]
            continue
            
        self.config.state_dict = state_dict
        LOG.logI("Reinterpret cast the model succeeded.")

    def loadStateDict(self):
        self.config.net = self.config.net.to(self.config.device)
        if not self.config.state_dict:
            LOG.logI("self.config.state_dict not initialized, omit loadStateDict()")
            return
        
        if self.config.model_reinterpret_cast:
            self.castStateDict()
            
        self.config.net.load_state_dict(self.config.state_dict, strict=False)
        self.config.net.eval()

    def loadJitModel(self):
        if not self.config.jit_model_path:
            LOG.logI("config.jit_model_path not specified, omit the loadJitModel")
            return

        if not self.config.is_forward_only:
            LOG.logI("You are in training mode, omit the loadJitModel")
            return

        self.config.net = torch.jit.load(self.config.jit_model_path, map_location=self.config.device)
        self.config.net.eval()

    def initLoader(self):
        #only init test_loader in base class
        if self.config.test_loader is None and self.config.is_forward_only:
            LOG.logE("You must set config.test_loader in config.py", exit=True)

    def initEMA(self):
        if self.config.ema is not True:
            return

        LOG.logI("Notice: You have enabled ema, which will increase the memory usage.")
        self.config.ema_updates = 0
        self.config.ema_net = copy.deepcopy(self.config.net)
        self.config.ema_net.to(self.config.device)
        if self.config.ema_decay is None:
            self.config.ema_decay = lambda x: 0.9999 * (1 - math.exp(-x / 2000))

        for p in self.config.ema_net.parameters():
            p.requires_grad_(False) 

    def init(self):
        #init self.config.device
        self.initDevice()
        #init self.config.net
        self.initNetWithCode()
        #init EMA
        self.initEMA()
        #init self.model_dict
        self.initStateDict()
        #just load model after audit
        self.loadStateDict()
        #jit load model
        self.loadJitModel()
        #just print model parameters info
        self._parametersInfo()
        #init loader
        self.initLoader()

    def export3rd(self, output_file=None):
        export3rd(self.config, output_file)

    def process(self):
        LOG.logE("You must reimplement process() to process config.sample", exit=True)

    def __call__(self, input=None):
        self.auditConfig()
        if self.config.sample is None and input is None:
            LOG.logE("either provide input at Deepvac(input), or set config.train.sample in your code.", exit=True)
        if input is not None:
            self.config.sample = input
        self.config.sample = self.config.sample.to(self.config.device)
        with torch.no_grad():
            x = self.process()
            #export 3rd
            self.export3rd()
        return x

#context manager for state switch
class deepvac_val_mode(object):
    def __init__(self, config):
        self.config = config
      
    def __enter__(self):
        self.config.is_train = False
        self.config.phase = 'VAL'
        self.config.dataset = self.config.val_dataset
        self.config.loader = self.config.val_loader
        self.config.batch_size = self.config.val_batch_size
        self.config.net.eval()
  
    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type is not None:
            return False
        self.config.is_train = True
        self.config.phase = 'TRAIN'
        self.config.dataset = self.config.train_dataset
        self.config.loader = self.config.train_loader
        self.config.batch_size = self.config.train_batch_size
        self.config.net.train()

#base class for train pipeline    
class DeepvacTrain(Deepvac):
    def __init__(self, deepvac_train_config):
        deepvac_train_config.is_forward_only=False
        super(DeepvacTrain, self).__init__(deepvac_train_config)
        self.initTrainContext()

    def auditConfig(self):
        #basic train config audit
        if not self.config.train_dataset:
            LOG.logE("You must set config.train_dataset in config.py",exit=True)
        if not self.config.val_dataset:
            LOG.logE("You must set config.val_dataset in config.py",exit=True)
        if not self.config.train_loader:
            LOG.logE("You must set config.train_loader in config.py",exit=True)
        if not self.config.val_loader:
            LOG.logE("You must set config.val_loader in config.py",exit=True)
        if not self.config.net:
            LOG.logE("You must set config.net in config.py",exit=True)
        if not self.config.criterion:
            LOG.logE("You must set config.criterion in config.py",exit=True)
        if not self.config.optimizer:
            LOG.logE("You must set config.optimizer in config.py",exit=True)
        
        #audit for amp
        if self.config.amp and self.config.device.type != 'cuda':
            LOG.logE("Error: amp can only be enabled when using cuda device", exit=True)

        #audit output dir
        if self.config.output_dir != 'output' and self.config.output_dir != './output':
            LOG.logW("According deepvac standard, you should set config.output_dir to [output] rather than [{}].".format(self.config.output_dir))

    def initTrainContext(self):
        self.config.epoch = 0
        self.config.iter = 0
        self.config.scaler = GradScaler()
        self.config.loader_time = AverageMeter()
        self.config.train_time = AverageMeter()
        self.initOutputDir()
        self.initSummaryWriter()
        self.initCriterion()
        self.initOptimizer()
        self.initScheduler()
        self.initCheckpoint()
        self.initTrainLoader()
        self.initValLoader()
        #must after initTrainLoader and initCheckpoint
        self.initEmaUpdates()

    def initOutputDir(self):
        self.config.output_dir = '{}/{}'.format(self.config.output_dir, self.config.branch)
        LOG.logI('model save dir: {}'.format(self.config.output_dir))
        #for DDP race condition
        os.makedirs(self.config.output_dir, exist_ok=True)

    def initSummaryWriter(self):
        event_dir = "{}/{}".format(self.config.log_dir, self.config.branch)
        self.config.writer = SummaryWriter(event_dir)
        if not self.config.tensorboard_port:
            return
        try:
            from tensorboard import program
        except Exception as e:
            LOG.logE(e.msg)
            LOG.logE("Make sure you have installed tensorboard with <pip install tensorboard>.", exit=True)
        tensorboard = program.TensorBoard()
        self.config.tensorboard_ip = '0.0.0.0' if self.config.tensorboard_ip is None else self.config.tensorboard_ip
        tensorboard.configure(argv=[None, '--host', str(self.config.tensorboard_ip),'--logdir', event_dir, "--port", str(self.config.tensorboard_port)])
        try:
            url = tensorboard.launch()
            LOG.logI('Tensorboard at {} '.format(url))
        except Exception as e:
            LOG.logE(e.msg, exit=True)

    def initCriterion(self):
        if self.config.criterion:
            LOG.logI("You configured criterion to {}".format(self.config.criterion))
            return
        self.config.criterion = torch.nn.CrossEntropyLoss()
        LOG.logW("You should set config.criterion in config.py, unless CrossEntropyLoss() is exactly what you need")

    def initCheckpoint(self):
        if not self.config.checkpoint_suffix or self.config.checkpoint_suffix == "":
            LOG.logI('Omit the checkpoint file since not specified...')
            return
        LOG.logI('Load checkpoint from {} folder'.format(self.config.output_dir))
        self.config.net.load_state_dict(torch.load(self.config.output_dir+'/model__{}'.format(self.config.checkpoint_suffix), map_location=self.config.device))
        state_dict = torch.load(self.config.output_dir+'/checkpoint__{}'.format(self.config.checkpoint_suffix), map_location=self.config.device)
        self.config.optimizer.load_state_dict(state_dict['optimizer'])
        self.config.epoch = state_dict['epoch']
        if self.config.scheduler:
            self.config.scheduler.load_state_dict(state_dict['scheduler'])
        
        if self.config.amp:
            self.initAmpScaler()

        if self.config.ema:
            self.config.ema_net.load_state_dict(state_dict['ema'])

    def initAmpScaler(self):
        LOG.logI("Will load scaler from checkpoint since you enabled amp, make sure the checkpoint was saved with amp enabled.")
        try:
            self.config.scaler.load_state_dict(state_dict["scaler"])
        except:
            LOG.logW("checkpoint was saved without amp enabled, so use fresh GradScaler instead.")
            self.config.scaler = GradScaler()

    def initScheduler(self):
        if self.config.scheduler:
            LOG.logI("You configured scheduler to {}".format(self.config.scheduler))
            return
        if not self.config.optimizer:
            LOG.logE("You must set config.optimizer or config.scheduler in config.py.",exit=True)
        if not self.config.lr_step:
            LOG.logE("You must set config.optimizer or config.lr_step in config.py.",exit=True)
        if isinstance(self.config.lr_step, list):
            self.config.scheduler = torch.optim.lr_scheduler.MultiStepLR(self.config.optimizer, self.config.lr_step,self.config.lr_factor)
        elif isinstance(self.config.lr_step, int):
            self.config.scheduler = torch.optim.lr_scheduler.StepLR(self.config.optimizer, self.config.lr_step,self.config.lr_factor)
        elif isinstance(self.config.lr_step, Callable):
            self.config.scheduler = torch.optim.lr_scheduler.LambdaLR(self.config.optimizer, lr_lambda=self.config.lr_step)
        else:
            LOG.logE("You should set config.scheduler in config.py.", exit=True)

    def initOptimizer(self):
        if self.config.optimizer:
            LOG.logI("You set optimizer to {} in config.py".format(self.config.optimizer))
            return
        self.initSgdOptimizer()
        LOG.logW("You should configure config.optimizer in config.py, unless SGD is exactly what you need")

    def initTrainLoader(self):
        if self.config.train_loader:
            LOG.logI("You set train_loader to {} in config.py".format(self.config.train_loader))
            return
        LOG.logE("You must set config.train_loader in config.py, or reimplement initTrainLoader() API in your DeepvacTrain subclass.")

    def initValLoader(self):
        if self.config.val_loader:
            LOG.logI("You set val_loader to {} in config.py".format(self.config.val_loader))
            return
        LOG.logE("You must set config.val_loader in config.py, or reimplement initValLoader() API in your DeepvacTrain subclass.")

    def initSgdOptimizer(self):
        self.config.optimizer = optim.SGD(self.config.net.parameters(),
            lr=self.config.lr,
            momentum=self.config.momentum,
            weight_decay=self.config.weight_decay,
            nesterov=self.config.nesterov
        )

    def initAdamOptimizer(self):
        self.config.optimizer = optim.Adam(
            self.config.net.parameters(),
            lr=self.config.lr,
            betas=self.config.betas if self.config.betas else (0.9, 0.999),
            weight_decay=self.config.weight_decay if self.config.weight_decay else 0
        )
        for group in self.config.optimizer.param_groups:
            group.setdefault('initial_lr', group['lr'])

    def initRmspropOptimizer(self):
        self.config.optimizer = optim.RMSprop(
            self.config.net.parameters(),
            lr=self.config.lr,
            momentum=self.config.momentum,
            weight_decay=self.config.weight_decay,
            # alpha=self.config.rmsprop_alpha,
            # centered=self.config.rmsprop_centered
        )

    def initEmaUpdates(self):
        if self.config.ema is not True:
            return
        self.config.ema_updates = self.config.epoch * len(self.config.train_loader) // self.config.nominal_batch_factor

    def addScalar(self, tag, value, step):
        self.config.writer.add_scalar(tag, value, step)

    def addImage(self, tag, image, step):
        self.config.writer.add_image(tag, image, step)

    @syszux_once
    def doGraph(self):
        if self.config.sample is None:
            LOG.logE("You must use doGraph in train() function.", exit=True)
        try:
            self.config.writer.add_graph(self.config.net, self.config.sample)
        except:
            LOG.logW("Tensorboard doGraph failed. You network foward may have more than one parameters?")
            LOG.logW("Seems you need to reimplement preIter function.")

    def updateEMA(self):
        if self.config.ema is not True:
            return
        self.config.ema_updates += 1
        d = self.config.ema_decay(self.config.ema_updates)
        msd = self.config.net.state_dict()
        with torch.no_grad():
            for k, v in self.config.ema_net.state_dict().items():
                if not v.is_floating_point():
                    continue
                v *= d
                v += (1. - d) * msd[k].detach()
    #For Deepvac user to reimplement
    def preIter(self):
        pass

    def doFeedData2Device(self):
        self.config.sample = self.config.sample.to(self.config.device)
        self.config.target = self.config.target.to(self.config.device)

    def postIter(self):
        pass

    #For Deepvac user to reimplement
    def preEpoch(self):
        pass

    #For Deepvac developer to reimplement
    def _preEpoch(self):
        pass

    #For Deepvac user to reimplement
    def postEpoch(self):
        pass

    #For Deepvac developer to reimplement
    def _postEpoch(self):
        pass

    def doForward(self):
        self.config.output = self.config.net(self.config.sample)

    def doLoss(self):
        self.config.loss = self.config.criterion(self.config.output, self.config.target)

    def doBackward(self):
        if self.config.amp:
            self.config.scaler.scale(self.config.loss).backward()
        else:
            self.config.loss.backward()

    def doOptimize(self):
        if self.config.iter % self.config.nominal_batch_factor != 0:
            return
        if self.config.amp:
            self.config.scaler.step(self.config.optimizer)
            self.config.scaler.update()
        else:
            self.config.optimizer.step()
        self.config.optimizer.zero_grad()

        if self.config.ema:
            self.updateEMA()

    def doSchedule(self):
        self.config.scheduler.step()

    def doLog(self):
        if self.config.step % self.config.log_every != 0:
            return
        self.addScalar('{}/Loss'.format(self.config.phase), self.config.loss.item(), self.config.iter)
        self.addScalar('{}/LoadDataTime(secs/batch)'.format(self.config.phase), self.config.loader_time.val, self.config.iter)
        self.addScalar('{}/TrainTime(secs/batch)'.format(self.config.phase), self.config.train_time.val, self.config.iter)
        LOG.logI('{}: [{}][{}/{}] [Loss:{}  Lr:{}]'.format(self.config.phase, self.config.epoch, self.config.step, self.config.loader_len,self.config.loss.item(),self.config.optimizer.param_groups[0]['lr']))

    def doSave(self, current_time):
        file_partial_name = '{}__acc_{}__epoch_{}__step_{}__lr_{}'.format(current_time, self.config.accuracy, self.config.epoch, self.config.step, self.config.optimizer.param_groups[0]['lr'])
        state_file = '{}/model__{}.pth'.format(self.config.output_dir, file_partial_name)
        checkpoint_file = '{}/checkpoint__{}.pth'.format(self.config.output_dir, file_partial_name)

        net = self.config.ema_net if self.config.ema else self.config.net
        torch.save(net.state_dict(), state_file)
        #save checkpoint
        torch.save({
            'optimizer': self.config.optimizer.state_dict(),
            'epoch': self.config.epoch,
            'scheduler': self.config.scheduler.state_dict() if self.config.scheduler else None,
            'ema': self.config.ema_net.state_dict() if self.config.ema else None,
            'scaler': self.config.scaler.state_dict() if self.config.amp else None},  checkpoint_file)
        #tensorboard
        self.addScalar('{}/Accuracy'.format(self.config.phase), self.config.accuracy, self.config.iter)

        #export 3rd
        self.export3rd(file_partial_name)

    def initTickTock(self):
        self.config.train_time.reset()
        self.config.loader_time.reset()

    def initStepAndSaveNumber(self):
        LOG.logI('Phase {} started...'.format(self.config.phase))
        self.config.loader_len = len(self.config.loader)
        save_every = self.config.loader_len//self.config.save_num
        save_list = list(range(0, self.config.loader_len + 1, save_every ))
        self.config.save_list = save_list[1:-1]
        LOG.logI('Model will be saved on step {} and the epoch end.'.format(self.config.save_list))
        self.addScalar('{}/LR'.format(self.config.phase), self.config.optimizer.param_groups[0]['lr'], self.config.epoch)

    def doTimekeeping(self):
        self.addScalar('{}/TrainTime(hours/epoch)'.format(self.config.phase), round(self.config.train_time.sum / 3600, 2), self.config.epoch)
        self.addScalar('{}/AverageBatchTrainTime(secs/epoch)'.format(self.config.phase), self.config.train_time.avg, self.config.epoch)
        self.addScalar('{}/AverageBatchLoadDataTime(secs/epoch)'.format(self.config.phase), self.config.loader_time.avg, self.config.epoch)

    def doIterTick(self):
        self.config.iter_tick = time.time()

    def doLoaderTock(self):
        self.config.loader_time.update(time.time() - self.config.iter_tick)

    def doTrainTock(self):
        self.config.train_time.update(time.time() - self.config.iter_tick)

    def doIterkeeping(self):
        self.config.iter += 1

    def train(self):
        self.initStepAndSaveNumber()
        self.initTickTock()
        self.preEpoch()
        self._preEpoch()
        self.doIterTick()
        for self.config.step, (self.config.sample, self.config.target) in enumerate(self.config.loader):
            self.doLoaderTock()
            self.preIter()
            self.doFeedData2Device()
            self.doGraph()
            with autocast(enabled=self.config.amp if self.config.amp else False):
                self.doForward()
                self.doLoss()
            self.doBackward()
            self.doOptimize()
            self.doLog()
            self.postIter()
            self.doIterkeeping()
            self.doTrainTock()
            if self.config.step in self.config.save_list:
                self.val()
            self.doIterTick()
        #epoch end
        self.postEpoch()
        self.doSchedule()
        self.doTimekeeping()

    def val(self, smoke=False):
        LOG.logI('Phase {} started...'.format(self.config.phase))
        with torch.no_grad(), deepvac_val_mode(self.config):
            self.preEpoch()
            for i, (self.config.sample, self.config.target) in enumerate(self.config.loader):
                self.preIter()
                self.doFeedData2Device()
                self.doForward()
                self.doLoss()
                self.postIter()
                if smoke:
                    self.export3rd()
                    return
                LOG.logI('{}: [{}][{}/{}]'.format(self.config.phase, self.config.epoch, i, len(self.config.loader)))
            self.postEpoch()
            self.doSave(getPrintTime())

    def processAccept(self):
        pass

    def process(self):
        self.config.optimizer.zero_grad()
        epoch_start = self.config.epoch
        for epoch in range(epoch_start, self.config.epoch_num):
            self.config.epoch = epoch
            LOG.logI('Epoch {} started...'.format(self.config.epoch))
            self.train()
            self.val()
            self.processAccept()

    def __call__(self):
        self.auditConfig()
        self.val(smoke=True)
        self.process()

class DeepvacDDP(DeepvacTrain):
    def __init__(self, deepvac_train_config):
        super(DeepvacDDP,self).__init__(deepvac_train_config)
        assert self.config.train_sampler is not None, "You should define config.train.train_sampler in config.py when training with DDP mode."

    def initDevice(self):
        super(DeepvacDDP, self).initDevice()
        parser = argparse.ArgumentParser(description='DeepvacDDP')
        parser.add_argument("--gpu", default=-1, type=int, help="gpu")
        parser.add_argument('--rank', default=-1, type=int, help='node rank for distributed training')
        self.config.args = parser.parse_args()
        self.config.map_location = {'cuda:%d' % 0: 'cuda:%d' % self.config.args.rank}
        #in DDP, device may come from command line
        if self.config.args.gpu:
            self.config.device = torch.device(self.config.args.gpu)
        torch.cuda.set_device(self.config.args.gpu)

    def initDDP(self):
        LOG.logI("Start dist.init_process_group {} {}@{} on {}".format(self.config.dist_url, self.config.args.rank, self.config.world_size - 1, self.config.args.gpu))
        dist.init_process_group(backend='nccl', init_method=self.config.dist_url, world_size=self.config.world_size, rank=self.config.args.rank)
        self.config.net = torch.nn.parallel.DistributedDataParallel(self.config.net, device_ids=[self.config.args.gpu])
        LOG.logI("Finish dist.init_process_group {} {}@{} on {}".format(self.config.dist_url, self.config.args.rank, self.config.world_size - 1, self.config.args.gpu))
        self.config.train_sampler = torch.utils.data.distributed.DistributedSampler(self.config.train_dataset)
        if self.config.train_loader:
            LOG.logW("Warning: config.train_loader will be override by Deepvac DDP is enabled. reimplement initDDP() API in your DeepvacTrain subclass to change this behaviour.")
        self.config.train_loader = DataLoader(
            self.config.train_dataset,
            batch_size=self.config.batch_size,
            num_workers=self.config.num_workers,
            shuffle=self.config.shuffle,
            sampler=self.config.train_sampler,
            collate_fn=self.config.collate_fn
        )

    def initTrainContext(self):
        self.initDDP()
        super(DeepvacDDP,self).initTrainContext()

    def initSummaryWriter(self):
        if self.config.args.rank != 0:
            return
        super(DeepvacDDP, self).initSummaryWriter()

    def _preEpoch(self):
        self.config.train_sampler.set_epoch(self.config.epoch)

    @syszux_once
    def smokeTest4Cast(self):
        if self.config.args.rank != 0:
            return
        super(DeepvacDDP, self).smokeTest4Cast()

    def doSave(self, time):
        if self.config.args.rank != 0:
            return
        super(DeepvacDDP, self).doSave(time)

    def addScalar(self, tag, value, step):
        if self.config.args.rank != 0:
            return
        super(DeepvacDDP, self).addScalar(tag, value, step)

    def addImage(self, tag, image, step):
        if self.config.args.rank != 0:
            return
        super(DeepvacDDP, self).addImage(tag, image, step)

    @syszux_once
    def doGraph(self):
        if self.config.args.rank != 0:
            return
        super(DeepvacDDP, self).doGraph()

if __name__ == "__main__":
    from config import config as deepvac_config
    vac = DeepvacTrain(deepvac_config.train)
    vac()



from PIL import Image

def get_image(image_path):
    # if the image_path is a remote url, read the image at first
    if isinstance(image_path, str):
        image = Image.open(image_path)
    else:
        image = image_path

    # convert non-RGB mode to RGB mode
    if image.mode != "RGB":
        image = image.convert("RGB")
    return image

def get_thumbnail(image):
    image.thumbnail((256, 256))
    return image

def get_colors(image_path):
    """ image instance
    """
    image = get_image(image_path)

    """ image thumbnail
        size: 256 * 256
        reduce the calculate time 
    """
    thumbnail = get_thumbnail(image)


    """ calculate the max colors the image cound have
        if the color is different in every pixel, the color counts may be the max.
        so : 
        max_colors = image.height * image.width
    """
    image_height = thumbnail.height
    image_width = thumbnail.width
    max_colors = image_height * image_width

    image_colors = image.getcolors(max_colors)
    return image_colors


def sort_by_rgb(colors_tuple):
    """ colors_tuple contains color count and color RGB
        we want to sort the tuple by RGB
        tuple[1]
    """
    sorted_tuple = sorted(colors_tuple, key=lambda x:x[1])
    return sorted_tuple

def rgb_maximum(colors_tuple):
    """ 
        colors_r max min
        colors_g max min
        colors_b max min

    """
    r_sorted_tuple = sorted(colors_tuple, key=lambda x:x[1][0])
    g_sorted_tuple = sorted(colors_tuple, key=lambda x:x[1][1])
    b_sorted_tuple = sorted(colors_tuple, key=lambda x:x[1][2])

    r_min = r_sorted_tuple[0][1][0]
    g_min = g_sorted_tuple[0][1][1]
    b_min = b_sorted_tuple[0][1][2]

    r_max = r_sorted_tuple[len(colors_tuple)-1][1][0]
    g_max = g_sorted_tuple[len(colors_tuple)-1][1][1]
    b_max = b_sorted_tuple[len(colors_tuple)-1][1][2]

    return {
        "r_max":r_max,
        "r_min":r_min,
        "g_max":g_max,
        "g_min":g_min,
        "b_max":b_max,
        "b_min":b_min,
        "r_dvalue":(r_max-r_min)/3,
        "g_dvalue":(g_max-g_min)/3,
        "b_dvalue":(b_max-b_min)/3
    }

def group_by_accuracy(sorted_tuple, accuracy=3):
    """ group the colors by the accuaracy was given
        the R G B colors will be depart to accuracy parts
        default accuracy = 3
        d_value = (max-min)/3
        [min, min+d_value), [min+d_value, min+d_value*2), [min+d_value*2, max)
    """
    rgb_maximum_json = rgb_maximum(sorted_tuple)
    r_min = rgb_maximum_json["r_min"]
    g_min = rgb_maximum_json["g_min"]
    b_min = rgb_maximum_json["b_min"]
    r_dvalue = rgb_maximum_json["r_dvalue"]
    g_dvalue = rgb_maximum_json["g_dvalue"]
    b_dvalue = rgb_maximum_json["b_dvalue"]

    rgb = [
            [[[], [], []], [[], [], []], [[], [], []]],
            [[[], [], []], [[], [], []], [[], [], []]],
            [[[], [], []], [[], [], []], [[], [], []]]
        ]

    for color_tuple in sorted_tuple:
        r_tmp_i = color_tuple[1][0]
        g_tmp_i = color_tuple[1][1]
        b_tmp_i = color_tuple[1][2]
        r_idx = 0 if r_tmp_i < (r_min+r_dvalue) else 1 if r_tmp_i < (r_min+r_dvalue*2) else 2
        g_idx = 0 if g_tmp_i < (g_min+g_dvalue) else 1 if g_tmp_i < (g_min+g_dvalue*2) else 2
        b_idx = 0 if b_tmp_i < (b_min+b_dvalue) else 1 if b_tmp_i < (b_min+b_dvalue*2) else 2
        rgb[r_idx][g_idx][b_idx].append(color_tuple)

    return rgb


def get_weighted_mean(grouped_image_color):
    """ calculate every group's weighted mean

        r_weighted_mean = sigma(r * count) / sigma(count)
        g_weighted_mean = sigma(g * count) / sigma(count)
        b_weighted_mean = sigma(b * count) / sigma(count)
    """
    sigma_count = 0
    sigma_r = 0
    sigma_g = 0
    sigma_b = 0

    for item in grouped_image_color:
        sigma_count += item[0]
        sigma_r += item[1][0] * item[0]
        sigma_g += item[1][1] * item[0]
        sigma_b += item[1][2] * item[0]

    r_weighted_mean = int(sigma_r / sigma_count)
    g_weighted_mean = int(sigma_g / sigma_count)
    b_weighted_mean = int(sigma_b / sigma_count)
    
    weighted_mean = (sigma_count, (r_weighted_mean, g_weighted_mean, b_weighted_mean))
    return weighted_mean

class Haishoku(object):

    """ init Haishoku obj
    """
    def __init__(self):
        self.dominant = None
        self.palette = None

    """ immediate api

        1. showPalette
        2. showDominant
        3. getDominant
        4. getPalette
    """
    def getColorsMean(image):
        # get colors tuple 
        image_colors = get_colors(image)

        # sort the image colors tuple
        sorted_image_colors = sort_by_rgb(image_colors)

        # group the colors by the accuaracy
        grouped_image_colors = group_by_accuracy(sorted_image_colors)

        # get the weighted mean of all colors
        colors_mean = []
        for i in range(3):
            for j in range(3):
                for k in range(3):
                    grouped_image_color = grouped_image_colors[i][j][k]
                    if 0 != len(grouped_image_color):
                        color_mean = get_weighted_mean(grouped_image_color)
                        colors_mean.append(color_mean)

        # return the most 8 colors
        temp_sorted_colors_mean = sorted(colors_mean)
        if 8 < len(temp_sorted_colors_mean):
            colors_mean = temp_sorted_colors_mean[len(temp_sorted_colors_mean)-8 : len(temp_sorted_colors_mean)]
        else:
            colors_mean = temp_sorted_colors_mean

        # sort the colors_mean
        colors_mean = sorted(colors_mean, reverse=True)

        return colors_mean
        
    def getDominant(image=None):
        # get the colors_mean
        colors_mean = Haishoku.getColorsMean(image)
        colors_mean = sorted(colors_mean, reverse=True)

        # get the dominant color
        dominant_tuple = colors_mean[0]
        dominant = dominant_tuple[1]
        return dominant


import torch
import torch.nn as nn
from .loss import LossBase

class Yolov5Loss(LossBase):
    def __init__(self, deepvac_config):
        super(Yolov5Loss, self).__init__(deepvac_config)
        self.cls = deepvac_config.cls
        self.box = deepvac_config.box
        self.obj = deepvac_config.obj
        self.device = deepvac_config.device
        self.strides = torch.Tensor(deepvac_config.strides)
        # define criteria
        BCEcls = nn.BCEWithLogitsLoss(pos_weight=torch.tensor([self.cls_pw], device=self.device))
        BCEobj = nn.BCEWithLogitsLoss(pos_weight=torch.tensor([self.obj_pw], device=self.device))
        if self.fl_gamma > 0:
            BCEcls, BCEobj = FocalLoss(BCEcls, self.fl_gamma), FocalLoss(BCEobj, self.fl_gamma)
        self.BCEcls, self.BCEobj = BCEcls, BCEobj
        # model info
        det = deepvac_config.model.detect
        self.ssi = (self.strides == 16).nonzero(as_tuple=False).item()
        for k in ("anchor_num", "class_num", "detect_layer_num", "anchors"):
            setattr(self, k, getattr(det, k))

    def auditConfig(self):
        self.gr = 1.0
        self.anchor_t = 4
        self.cls_pw = 1.0
        self.obj_pw = 1.0
        self.fl_gamma = 0.0
        self.autobalance = False
        self.balance = [4.0, 1.0, 0.4]
        self.cp, self.cn = self.smoothBCE(eps=0.0)

    def build_target(self, pred, target):
        target_num = target.size(0)
        tcls, tbox, indices, anch = [], [], [], []
        # normalized to gridspace gain
        gain = torch.ones(7, device=self.device)
        # targets: (N, 6) -> (3, N, 7), at last append anchor index
        target = [torch.cat([target, i.repeat(target_num).view(target_num, 1)], dim=1) for i in torch.arange(self.anchor_num, device=self.device)]
        target = torch.stack(target, dim=0)

        off = torch.tensor([[0, 0], [1, 0], [0, 1], [-1, 0], [0, -1]], device=self.device) * 0.5
        for i in range(self.detect_layer_num):
            anchors = self.anchors[i]
            # pred: [(n, c, h1, w1, 9), (n, c, h2, w2, 9), (n, c, h3, w3, 9)] gain: [1, 1, w, h, w, h, 1]
            gain[2:6] = torch.tensor(pred[i].shape)[[3, 2, 3, 2]]
            # target: [[[id, cls, cx, cy, w, h, anchor_id], ...n...]]
            # normalize by w, h -> origin size
            t = target * gain
            if target_num:
                # anchors: (3, 2) -> (3, 1, 2)
                wh_radio = t[:, :, 4:6] / anchors.unsqueeze(dim=1)
                # index: (3, N) t: (3, N, 7) -> (n, 7)
                index = torch.max(wh_radio, 1/wh_radio).max(dim=2)[0] < self.anchor_t
                t = t[index]
                # cxcy: [[cx, xy], ...n...]
                cxcy = t[:, 2:4]
                # inverse_cxcy: [[w-cx, h-cy], ...n...]
                inverse_cxcy = gain[[2, 3]] - cxcy
                # cx_index: x_index  cy_index: y_index
                cx_index, cy_index = ((cxcy.fmod(1.) < 0.5) & (cxcy > 1.)).T
                # inverse_cx_index: x_index  inverse_cy_index: y_index
                inverse_cx_index, inverse_cy_index = ((inverse_cxcy % 1. < 0.5) & (inverse_cxcy > 1.)).T
                # cx_index: (n) -> (5, n)
                cx_index = torch.stack((torch.ones_like(cx_index), cx_index, cy_index, inverse_cx_index, inverse_cy_index))
                # t: (n, 7) -> (5, n, 7) -> (n', 7)
                t = t.unsqueeze(dim=0).repeat((5, 1, 1))[cx_index]
                offsets = (torch.zeros_like(cxcy.unsqueeze(dim=0)) + off.unsqueeze(dim=1))[cx_index]
                # offsets = (torch.zeros_like(cxcy)[None] + off[:, None])[cx_index]
            else:
                t = target[0]
                offsets = 0

            img_id, cls = t[:, :2].long().T
            cxcy = t[:, 2:4]
            wh = t[:, 4:6]
            cxcy_index = (cxcy - offsets).long()
            cx_index, cy_index = cxcy_index.T

            anchor_index = t[:, 6].long()
            indices.append((img_id, anchor_index, cy_index.clamp_(0, gain[3] - 1), cx_index.clamp_(0, gain[2] - 1)))
            tbox.append(torch.cat((cxcy - cxcy_index, wh), 1))
            anch.append(anchors[anchor_index])
            tcls.append(cls)
        return tcls, tbox, indices, anch

    def compute_loss(self, p, tcls, tbox, indices, anchors, balance):
        img_id, anchor_index, cy_index, cx_index = indices
        tobj = torch.zeros_like(p[..., 0], device=self.device)
        target_num = img_id.size(0)
        if not target_num:
            lobj = self.BCEobj(p[..., 4], tobj) * balance
            balance = balance * 0.9999 + 0.0001 / lobj.detach().item() if self.autobalance else balance
            return 0, 0, lobj
        # p: [px, py, pw, ph, conf, cls] ...
        ps = p[img_id, anchor_index, cy_index, cx_index]
        # Regression
        pcxcy = ps[:, :2].sigmoid() * 2. - 0.5
        pwh = (ps[:, 2:4].sigmoid() * 2) ** 2 * anchors
        pbox = torch.cat((pcxcy, pwh), 1)
        iou = self.bbox_iou(pbox.T, tbox, x1y1x2y2=False, CIou=True)
        lbox = (1.0 - iou).mean()
        # Objectness
        tobj[img_id, anchor_index, cy_index, cx_index] = (1.0 - self.gr) + self.gr * iou.detach().clamp(0).type(tobj.dtype)
        lobj = self.BCEobj(p[..., 4], tobj) * balance
        balance = balance * 0.9999 + 0.0001 / lobj.detach().item() if self.autobalance else balance
        # Classification
        if self.class_num <= 1:
            return lbox, 0, lobj
        t = torch.full_like(ps[:, 5:], self.cn, device=self.device)
        t[range(target_num), tcls] = self.cp
        lcls = self.BCEcls(ps[:, 5:], t)
        return lbox, lcls, lobj

    def __call__(self, pred, target):
        lcls, lbox, lobj = torch.zeros(1, device=self.device), torch.zeros(1, device=self.device), torch.zeros(1, device=self.device)
        tcls, tbox, indices, anchors = self.build_target(pred, target)
        # pred: [(n, c, h1, w1, 9), (n, c, h2, w2, 9), (n, c, h3, w3, 9)]
        for i, p in enumerate(pred):
            libox, licls, liobj = self.compute_loss(p, tcls[i], tbox[i], indices[i], anchors[i], self.balance[i])
            lbox += libox
            lcls += licls
            lobj += liobj
        self.balance = [x / self.balance[self.ssi] for x in self.balance] if self.autobalance else self.balance
        lbox *= self.box
        lobj *= self.obj
        lcls *= self.cls
        bs = pred[0].size(0)
        loss = lbox + lobj + lcls
        return loss * bs, torch.cat((lbox, lobj, lcls, loss)).detach()

    @staticmethod
    def bbox_iou(box1, box2, x1y1x2y2=True, GIou=False, DIou=False, CIou=False, eps=1e-9):
        box2 = box2.T
        if x1y1x2y2:
            b1_x1, b1_y1, b1_x2, b1_y2 = box1[0], box1[1], box1[2], box1[3]
            b2_x1, b2_y1, b2_x2, b2_y2 = box2[0], box2[1], box2[2], box2[3]
        else:
            b1_x1, b1_x2 = box1[0] - box1[2] / 2, box1[0] + box1[2] / 2
            b1_y1, b1_y2 = box1[1] - box1[3] / 2, box1[1] + box1[3] / 2
            b2_x1, b2_x2 = box2[0] - box2[2] / 2, box2[0] + box2[2] / 2
            b2_y1, b2_y2 = box2[1] - box2[3] / 2, box2[1] + box2[3] / 2
        inter = (torch.min(b1_x2, b2_x2) - torch.max(b1_x1, b2_x1)).clamp(0) * \
                (torch.min(b1_y2, b2_y2) - torch.max(b1_y1, b2_y1)).clamp(0)
        coords = (b1_x1, b1_x2, b1_y1, b1_y2, b2_x1, b2_x2, b2_y1, b2_y2)
        w1, h1 = b1_x2 - b1_x1, b1_y2 - b1_y1 + eps
        w2, h2 = b2_x2 - b2_x1, b2_y2 - b2_y1 + eps
        whs = (w1, h1, w2, h2)
        union = w1 * h1 + w2 * h2 - inter + eps
        iou = inter / union

        if not GIou and not DIou and not CIou:
            return iou
        elif GIou:
            return Yolov5Loss.compute_GIou(iou, coords, union)[-1]
        elif DIou:
            return Yolov5Loss.compute_DIou(iou, coords, union)[-1]
        elif CIou:
            return Yolov5Loss.compute_CIou(iou, coords, whs, union)

    @staticmethod
    def compute_GIou(iou, coords, union, eps=1e-9):
        b1_x1, b1_x2, b1_y1, b1_y2, b2_x1, b2_x2, b2_y1, b2_y2 = coords
        cw = torch.max(b1_x2, b2_x2) - torch.min(b1_x1, b2_x1)
        ch = torch.max(b1_y2, b2_y2) - torch.min(b1_y1, b2_y1)
        c_area = cw * ch + eps
        giou = iou - (c_area - union) / c_area
        return cw, ch, giou

    @staticmethod
    def compute_DIou(iou, coords, union, eps=1e-9):
        cw, ch, _ = Yolov5Loss.compute_GIou(iou, coords, union)
        c2 = cw ** 2 + ch ** 2 + eps
        b1_x1, b1_x2, b1_y1, b1_y2, b2_x1, b2_x2, b2_y1, b2_y2 = coords
        rho2 = ((b2_x1 + b2_x2 - b1_x1 - b1_x2) ** 2 + (b2_y1 + b2_y2 - b1_y1 - b1_y2) ** 2) / 4
        diou = iou - rho2 / c2
        return c2, rho2, diou

    @staticmethod
    def compute_CIou(iou, coords, whs, union, eps=1e-9):
        cw, ch, _ = Yolov5Loss.compute_GIou(iou, coords, union)
        c2, rho2, _ = Yolov5Loss.compute_DIou(iou, coords, union)
        w1, h1, w2, h2 = whs
        v = (4 / math.pi ** 2) * torch.pow(torch.atan(w2 / h2) - torch.atan(w1 / h1), 2)
        with torch.no_grad():
            alpha = v / ((1 + eps) - iou + v)
        ciou = iou - (rho2 / c2 + v * alpha)
        return ciou

    @staticmethod
    def smoothBCE(eps=0.1):
        return 1.0 - 0.5 * eps, 0.5 * eps


import torch.nn as nn

class hswish(nn.Module):
    def __init__(self, inplace=True):
        super(hswish, self).__init__()
        self.relu = nn.ReLU6(inplace=inplace)

    def forward(self, x):
        out = x * self.relu(x+3) / 6
        return out

class hsigmoid(nn.Module):
    def __init__(self, inplace=True):
        super(hsigmoid, self).__init__()
        self.relu = nn.ReLU6(inplace=inplace)

    def forward(self, x):
        out = self.relu(x+3) / 6
        return out


import cv2

class Remaper(object):
    def __init__(self):
        self.period = 360  
        self.min = 1
        self.max = 5

    def apply(self, word_img, text_box_pnts):
        """
        :param word_img:  word image with big background
        :param text_box_pnts: left-top, right-top, right-bottom, left-bottom of text word
        :return:
        """
        max_val = np.random.uniform(self.min, self.max)

        h = word_img.shape[0]
        w = word_img.shape[1]

        img_x = np.zeros((h, w), np.float32)
        img_y = np.zeros((h, w), np.float32)

        xmin = text_box_pnts[0][0]
        xmax = text_box_pnts[1][0]
        ymin = text_box_pnts[0][1]
        ymax = text_box_pnts[2][1]

        remap_y_min = ymin
        remap_y_max = ymax

        for y in range(h):
            for x in range(w):
                remaped_y = y + self._remap_y(x, max_val)

                if y == ymin:
                    if remaped_y < remap_y_min:
                        remap_y_min = remaped_y

                if y == ymax:
                    if remaped_y > remap_y_max:
                        remap_y_max = remaped_y

                # 某一个位置的 y 值应该为哪个位置的 y 值
                img_y[y, x] = remaped_y
                # 某一个位置的 x 值应该为哪个位置的 x 值
                img_x[y, x] = x

        remaped_text_box_pnts = [
            [xmin, remap_y_min],
            [xmax, remap_y_min],
            [xmax, remap_y_max],
            [xmin, remap_y_max]
        ]

        # TODO: use cuda::remap
        dst = cv2.remap(word_img, img_x, img_y, cv2.INTER_CUBIC)
        return dst, remaped_text_box_pnts

    def _remap_y(self, x, max_val):
        return int(max_val * np.math.sin(2 * 3.14 * x / self.period))



from collections import OrderedDict
import torch.nn as nn
from .activate_layer import hswish

class Conv2dBN(nn.Sequential):
    def __init__(self, in_planes, out_planes, kernel_size=3, stride=1, padding=None, groups=1):
        if padding is None:
            padding = (kernel_size - 1) // 2
        super(Conv2dBN, self).__init__(
            nn.Conv2d(in_planes, out_planes, kernel_size, stride, padding, bias=False),
            nn.BatchNorm2d(out_planes)
        )

class Conv2dBNWithName(nn.Sequential):
    def __init__(self, in_planes, out_planes, kernel_size=3, stride=1, padding=None, groups=1):
        if padding is None:
            padding = (kernel_size - 1) // 2
        super(Conv2dBNWithName, self).__init__(OrderedDict([
            ('conv', nn.Conv2d(in_planes, out_planes, kernel_size, stride, padding, groups=groups, bias=False)),
            ('bn', nn.BatchNorm2d(out_planes))
        ]))
        
class Conv2dBNReLU(nn.Sequential):
    def __init__(self, in_planes, out_planes, kernel_size=3, stride=1, padding=None, groups=1):
        if padding is None:
            padding = (kernel_size - 1) // 2
        super(Conv2dBNReLU, self).__init__(
            nn.Conv2d(in_planes, out_planes, kernel_size, stride, padding, groups=groups, bias=False),
            nn.BatchNorm2d(out_planes, momentum=0.1),
            nn.ReLU(inplace=True)
        )

class Conv2dBNPReLU(nn.Sequential):
    def __init__(self, in_planes, out_planes, kernel_size=3, stride=1, padding=None, groups=1):
        if padding is None:
            padding = (kernel_size - 1) // 2
        super(Conv2dBNPReLU, self).__init__(
            nn.Conv2d(in_planes, out_planes, kernel_size, stride, padding, groups=groups, bias=False),
            nn.BatchNorm2d(out_planes, momentum=0.1),
            nn.PReLU(out_planes)
        )

class Conv2dBnAct(nn.Sequential):
    def __init__(self, in_planes, out_planes, kernel_size=3, stride=1, padding=None, groups=1, act=True):
        if padding is None:
            padding = (kernel_size - 1) // 2
        super(Conv2dBnAct, self).__init__(
            nn.Conv2d(in_planes, out_planes, kernel_size, stride, padding, groups=groups, bias=False),
            nn.BatchNorm2d(out_planes),
            nn.SiLU() if (act is True) else (act if isinstance(act, nn.Module) else nn.Identity())
        )

class Conv2dBNHardswish(nn.Sequential):
    def __init__(self, in_planes, out_planes, kernel_size=3, stride=1, padding=None, groups=1):
        if padding is None:
            padding = (kernel_size - 1) // 2
        super(Conv2dBNHardswish, self).__init__(
            nn.Conv2d(in_planes, out_planes, kernel_size, stride, padding, groups=groups, bias=False),
            nn.BatchNorm2d(out_planes, momentum=0.1),
            nn.Hardswish()
        )

class Conv2dBNHswish(nn.Sequential):
    def __init__(self, in_planes, out_planes, kernel_size=3, stride=1, padding=None, groups=1):
        if padding is None:
            padding = (kernel_size - 1) // 2
        super(Conv2dBNHswish, self).__init__(
            nn.Conv2d(in_planes, out_planes, kernel_size, stride, padding, groups=groups, bias=False),
            nn.BatchNorm2d(out_planes, momentum=0.1),
            hswish(inplace=True)
        )

class DepthWiseConv2d(nn.Module):
    def __init__(self, inplanes: int, outplanes: int, kernel_size: int, stride: int, padding: int, groups: int, residual: bool=False):
        super(DepthWiseConv2d, self).__init__()
        self.conv1 = Conv2dBNPReLU(inplanes, groups, kernel_size=1, stride=1, padding=0)
        self.conv2 = Conv2dBNPReLU(groups, groups, kernel_size, stride, padding, groups)
        self.conv3 = nn.Conv2d(groups, outplanes, kernel_size=1, stride=1, padding=0, bias=False)
        self.bn = nn.BatchNorm2d(outplanes)
        self.residual = residual

    def forward(self, x):
        identity = x
        x = self.conv1(x)
        x = self.conv2(x)
        x = self.conv3(x)
        x = self.bn(x)
        if self.residual:
            x = identity + x
        return x

class Conv2dBNLeakyReLU(nn.Sequential):
    def __init__(self, in_planes, out_planes, kernel_size=3, stride=1, padding=None, groups=1, leaky=0):
        if padding is None:
            padding = (kernel_size - 1) // 2
        #leaky = 0.1 if out_planes <= 64 else 0
        super(Conv2dBNLeakyReLU, self).__init__(
            nn.Conv2d(in_planes, out_planes, kernel_size, stride, padding, bias=False),
            nn.BatchNorm2d(out_planes),
            nn.LeakyReLU(negative_slope=leaky, inplace=True)
        )


from ..utils import LOG
from .base import DeepvacCast, FXQuantizeAndScript

class ScriptCast(DeepvacCast):
    def __init__(self, deepvac_train_config):
        super(ScriptCast,self).__init__(deepvac_train_config)

    def auditConfig(self):
        if not self.config.script_model_dir:
            return False

        if not self.config.static_quantize_dir:
            return True
            
        if self.config.is_forward_only and self.config.test_loader is None:
            LOG.logE("You must set config.train.test_loader in config.py when config.train.static_quantize_dir is enabled.", exit=True)
        if not self.config.is_forward_only and self.config.val_loader is None:
            LOG.logE("You must set config.train.val_loader in config.py when config.train.static_quantize_dir is enabled.", exit=True)

        return True

    def process(self, cast_output_file=None):
        output_script_file = self.config.script_model_dir
        if cast_output_file:
            output_script_file = '{}/script__{}.pt'.format(self.config.output_dir, cast_output_file)

        LOG.logI("config.script_model_dir found, save script model to {}...".format(output_script_file))

        net = self.getNet()
        script_model = FXQuantizeAndScript(self.config, output_script_file)
        script_model.save()

        if self.config.dynamic_quantize_dir:
            LOG.logI("You have enabled config.dynamic_quantize_dir, will dynamic quantize the model...")
            script_model.saveDQ()

        if self.config.static_quantize_dir:
            LOG.logI("You have enabled config.static_quantize_dir, will static quantize the model...")
            script_model.saveSQ()


from ..utils import LOG
from .base import DeepvacCast

class OnnxCast(DeepvacCast):
    def __init__(self, deepvac_train_config):
        super(OnnxCast,self).__init__(deepvac_train_config)

    def auditConfig(self):
        if not self.config.onnx_model_dir:
            return False
        return True

    def process(self, cast_output_file=None):
        output_onnx_file = self.config.onnx_model_dir
        if cast_output_file:
            output_onnx_file = '{}/onnx__{}.onnx'.format(self.config.output_dir, cast_output_file)
            self.config.onnx_model_dir = output_onnx_file
        LOG.logI("config.onnx_model_dir found, save onnx model to {}...".format(output_onnx_file))
        self.exportOnnx()


import torch
from torch.quantization.fuser_method_mappings import DEFAULT_OP_LIST_TO_FUSER_METHOD
from torch.quantization import QuantStub, DeQuantStub

def get_module(model, submodule_key):
    tokens = submodule_key.split('.')
    cur_mod = model
    for s in tokens:
        cur_mod = getattr(cur_mod, s)
    return cur_mod

def set_module(model, submodule_key, module):
    tokens = submodule_key.split('.')
    sub_tokens = tokens[:-1]
    cur_mod = model
    for s in sub_tokens:
        cur_mod = getattr(cur_mod, s)
    setattr(cur_mod, tokens[-1], module)

def get_fuser_module_index(mod_list):
    rc = []
    if len(mod_list) < 2:
        return rc
    keys = sorted(list(DEFAULT_OP_LIST_TO_FUSER_METHOD.keys()), key=lambda x: len(x), reverse=True)
    mod2fused_list = [list(x) for x in keys]

    for mod2fused in mod2fused_list:
        if len(mod2fused) > len(mod_list):
            continue
        mod2fused_idx = [(i, i+len(mod2fused)) for i in range(len(mod_list) - len(mod2fused) + 1) if mod_list[i:i+len(mod2fused)] == mod2fused]
        if not mod2fused_idx:
            continue

        for idx in mod2fused_idx:
            start,end = idx
            mod_list[start: end] = [None] * len(mod2fused)

        rc.extend(mod2fused_idx)

    return rc

def auto_fuse_model(model):
    module_names = []
    module_types = []
    for name, m in model.named_modules():
        module_names.append(name)
        module_types.append(type(m))

    if len(module_types) < 2:
        return model

    module_idxs = get_fuser_module_index(module_types)
    modules_to_fuse = [module_names[mi[0]:mi[1]] for mi in module_idxs]
    new_model = torch.quantization.fuse_modules(model, modules_to_fuse)
    return new_model

class DeQuantStub(nn.Module):
    def __init__(self):
        super(DeQuantStub, self).__init__()

    def forward(self, x: Any) -> Any:
        return x

def calibrate(model, data_loader):
    model.eval()
    with torch.no_grad():
        for sample, target in data_loader:
            model(sample)

def prepareQAT(deepvac_train_config):
    if not deepvac_train_config.qat_dir:
        return

    if deepvac_train_config.is_forward_only:
        LOG.logI("You are in forward_only mode, omit the parepareQAT()")
        return

    LOG.logI("You have enabled QAT, this step is only for prepare.")

    if deepvac_train_config.qat_net_prepared:
        LOG.logE("Error: You have already prepared the model for QAT.", exit=True)

    backend = 'fbgemm'
    if deepvac_train_config.quantize_backend:
        backend = deepvac_train_config.quantize_backend
    deepvac_train_config.qat_net_prepared = DeepvacQAT(deepvac_train_config.net).to(deepvac_train_config.device)
    deepvac_train_config.qat_net_prepared.qconfig = torch.quantization.get_default_qat_qconfig(backend)
    torch.quantization.prepare_qat(deepvac_train_config.qat_net_prepared, inplace=True)
    #after this, train net will be transfered to QAT !
    deepvac_train_config.net = deepvac_train_config.qat_net_prepared

class DeepvacQAT(torch.nn.Module):
    def __init__(self, net2qat):
        super(DeepvacQAT, self).__init__()
        self.quant = QuantStub()
        self.net2qat = auto_fuse_model(net2qat)
        self.dequant = DeQuantStub()

    def forward(self, x):
        x = self.quant(x)
        x = self.net2qat(x)
        x = self.dequant(x)
        return x


import setuptools
import sys
def readme():
    with open('README.md') as f:
        return f.read()

version="0.5.0"
setuptools.setup(name='deepvac',
    version=version,
    description='PyTorch python project standard',
    long_description='PyTorch python project standard',
    keywords='PyTorch python project standard',
    url='https://github.com/DeepVAC/deepvac',
    download_url="",
    author='Gemfield',
    author_email='gemfield@civilnet.cn',
    packages=setuptools.find_packages(),
    include_package_data=True,
    package_dir={
        'deepvac': 'deepvac',
    },

    install_requires=[
        'opencv_python',
        'numpy',
        'Pillow',
        'scipy',
        'tensorboard'
    ],
    classifiers  = [
        "Programming Language :: Python :: 3",
        "Topic :: Software Development :: Libraries"
     ],
    zip_safe=False)



import torch.nn as nn
import torch.nn.functional as F
from .conv_layer import Conv2dBNLeakyReLU

class FPN(nn.Module):
    def __init__(self, in_planes: list, out_planes: int):
        super(FPN,self).__init__()

        leaky = 0.1 if out_planes <= 64 else 0
        self.conv1 = Conv2dBNLeakyReLU(in_planes[0], out_planes, kernel_size=1, padding=0, leaky=leaky)
        self.conv2 = Conv2dBNLeakyReLU(in_planes[1], out_planes, kernel_size=1, padding=0, leaky=leaky)
        self.conv3 = Conv2dBNLeakyReLU(in_planes[2], out_planes, kernel_size=1, padding=0, leaky=leaky)

        self.conv4 = Conv2dBNLeakyReLU(out_planes, out_planes, padding=1, leaky=leaky)
        self.conv5 = Conv2dBNLeakyReLU(out_planes, out_planes, padding=1, leaky=leaky)

    def forward(self, input):
        output1 = self.conv1(input[0])
        output2 = self.conv2(input[1])
        output3 = self.conv3(input[2])

        up3 = F.interpolate(output3, size=[output2.size(2), output2.size(3)], mode="nearest")
        output2 = output2 + up3
        output2 = self.conv5(output2)

        up2 = F.interpolate(output2, size=[output1.size(2), output1.size(3)], mode="nearest")
        output1 = output1 + up2
        output1 = self.conv4(output1)
        out = [output1, output2, output3]
        return out


import torch
import torch.nn as nn
import torch.nn.functional as F
from .loss import LossBase

class MultiBoxLoss(nn.Module):
    def __init__(self, num_classes, overlap_thresh, prior_for_matching, bkg_label, neg_mining, neg_pos, neg_overlap, encode_target, device='cpu'):
        super(MultiBoxLoss, self).__init__()
        self.num_classes = num_classes
        self.threshold = overlap_thresh
        self.background_label = bkg_label
        self.encode_target = encode_target
        self.use_prior_for_matching = prior_for_matching
        self.do_neg_mining = neg_mining
        self.negpos_ratio = neg_pos
        self.neg_overlap = neg_overlap
        self.variance = [0.1, 0.2]
        self.device = device

    def _match(self, threshold, truths, priors, variances, labels, landms, loc_t, conf_t, landm_t, idx):
        overlaps = self._jaccard(
            truths,
            self._pointForm(priors)
        )
        # (Bipartite Matching)
        # [1,num_objects] best prior for each ground truth
        best_prior_overlap, best_prior_idx = overlaps.max(1, keepdim=True)

        # ignore hard gt
        valid_gt_idx = best_prior_overlap[:, 0] >= 0.2
        best_prior_idx_filter = best_prior_idx[valid_gt_idx, :]
        if best_prior_idx_filter.shape[0] <= 0:
            loc_t[idx] = 0
            conf_t[idx] = 0
            return

        # [1,num_priors] best ground truth for each prior
        best_truth_overlap, best_truth_idx = overlaps.max(0, keepdim=True)
        best_truth_idx.squeeze_(0)
        best_truth_overlap.squeeze_(0)
        best_prior_idx.squeeze_(1)
        best_prior_idx_filter.squeeze_(1)
        best_prior_overlap.squeeze_(1)
        best_truth_overlap.index_fill_(0, best_prior_idx_filter, 2)  # ensure best prior
        # ensure every gt matches with its prior of max overlap
        for j in range(best_prior_idx.size(0)):
            best_truth_idx[best_prior_idx[j]] = j

        matches = truths[best_truth_idx]
        conf = labels[best_truth_idx]
        conf[best_truth_overlap < threshold] = 0
        loc = self._encode(matches, priors, variances)

        matches_landm = landms[best_truth_idx]
        landm = self._encodeLandm(matches_landm, priors, variances)
        loc_t[idx] = loc    # [num_priors,4] encoded offsets to learn
        conf_t[idx] = conf  # [num_priors] top class label for each prior
        landm_t[idx] = landm

    def _jaccard(self, box_a, box_b):
        inter = self._intersect(box_a, box_b)
        area_a = ((box_a[:, 2]-box_a[:, 0]) *
                (box_a[:, 3]-box_a[:, 1])).unsqueeze(1).expand_as(inter)  # [A,B]
        area_b = ((box_b[:, 2]-box_b[:, 0]) *
                (box_b[:, 3]-box_b[:, 1])).unsqueeze(0).expand_as(inter)  # [A,B]
        union = area_a + area_b - inter
        return inter / union  # [A,B]

    def _pointForm(self, boxes):
        return torch.cat((boxes[:, :2] - boxes[:, 2:]/2,     # xmin, ymin
                boxes[:, :2] + boxes[:, 2:]/2), 1)  # xmax, ymax

    def _intersect(self, box_a, box_b):
        A = box_a.size(0)
        B = box_b.size(0)
        max_xy = torch.min(box_a[:, 2:].unsqueeze(1).expand(A, B, 2),
                box_b[:, 2:].unsqueeze(0).expand(A, B, 2))
        min_xy = torch.max(box_a[:, :2].unsqueeze(1).expand(A, B, 2),
                box_b[:, :2].unsqueeze(0).expand(A, B, 2))
        inter = torch.clamp((max_xy - min_xy), min=0)
        return inter[:, :, 0] * inter[:, :, 1]

    def _encode(self, matched, priors, variances):
        # dist b/t match center and prior's center
        g_cxcy = (matched[:, :2] + matched[:, 2:])/2 - priors[:, :2]
        # encode variance
        g_cxcy /= (variances[0] * priors[:, 2:])
        # match wh / prior wh
        g_wh = (matched[:, 2:] - matched[:, :2]) / priors[:, 2:]
        g_wh = torch.log(g_wh) / variances[1]
        # return target for smooth_l1_loss
        return torch.cat([g_cxcy, g_wh], 1)  # [num_priors,4]

    def _encodeLandm(self, matched, priors, variances):
        # dist b/t match center and prior's center
        matched = torch.reshape(matched, (matched.size(0), 5, 2))
        priors_cx = priors[:, 0].unsqueeze(1).expand(matched.size(0), 5).unsqueeze(2)
        priors_cy = priors[:, 1].unsqueeze(1).expand(matched.size(0), 5).unsqueeze(2)
        priors_w = priors[:, 2].unsqueeze(1).expand(matched.size(0), 5).unsqueeze(2)
        priors_h = priors[:, 3].unsqueeze(1).expand(matched.size(0), 5).unsqueeze(2)
        priors = torch.cat([priors_cx, priors_cy, priors_w, priors_h], dim=2)
        g_cxcy = matched[:, :, :2] - priors[:, :, :2]
        # encode variance
        g_cxcy /= (variances[0] * priors[:, :, 2:])
        # g_cxcy /= priors[:, :, 2:]
        g_cxcy = g_cxcy.reshape(g_cxcy.size(0), -1)
        # return target for smooth_l1_loss
        return g_cxcy

    def _logSumExp(self, x):
        x_max = x.data.max()
        return torch.log(torch.sum(torch.exp(x-x_max), 1, keepdim=True)) + x_max

    def forward(self, predictions, priors, targets):
        loc_data, conf_data, landm_data = predictions
        priors = priors
        num = loc_data.size(0)
        num_priors = (priors.size(0))

        # match priors (default boxes) and ground truth boxes
        loc_t = torch.Tensor(num, num_priors, 4)
        landm_t = torch.Tensor(num, num_priors, 10)
        conf_t = torch.LongTensor(num, num_priors)

        for idx in range(num):
            truths = targets[idx][:, :4].data
            labels = targets[idx][:, -1].data
            landms = targets[idx][:, 4:14].data
            defaults = priors.data
            self._match(self.threshold, truths, defaults, self.variance, labels, landms, loc_t, conf_t, landm_t, idx)

        loc_t = loc_t.to(self.device)
        conf_t = conf_t.to(self.device)
        landm_t = landm_t.to(self.device)

        zeros = torch.tensor(0).cuda()
        # landm Loss (Smooth L1)
        # Shape: [batch,num_priors,10]
        pos1 = conf_t > zeros
        num_pos_landm = pos1.long().sum(1, keepdim=True)
        N1 = max(num_pos_landm.data.sum().float(), 1)
        pos_idx1 = pos1.unsqueeze(pos1.dim()).expand_as(landm_data)
        landm_p = landm_data[pos_idx1].view(-1, 10)
        landm_t = landm_t[pos_idx1].view(-1, 10)
        loss_landm = F.smooth_l1_loss(landm_p, landm_t, reduction='sum')

        pos = conf_t != zeros
        conf_t[pos] = 1

        # Localization Loss (Smooth L1)
        # Shape: [batch,num_priors,4]
        pos_idx = pos.unsqueeze(pos.dim()).expand_as(loc_data)
        loc_p = loc_data[pos_idx].view(-1, 4)
        loc_t = loc_t[pos_idx].view(-1, 4)
        loss_l = F.smooth_l1_loss(loc_p, loc_t, reduction='sum')

        # Compute max conf across batch for hard negative mining
        batch_conf = conf_data.view(-1, self.num_classes)
        loss_c = self._logSumExp(batch_conf) - batch_conf.gather(1, conf_t.view(-1, 1))

        # Hard Negative Mining
        loss_c[pos.view(-1, 1)] = 0 # filter out pos boxes for now
        loss_c = loss_c.view(num, -1)
        _, loss_idx = loss_c.sort(1, descending=True)
        _, idx_rank = loss_idx.sort(1)
        num_pos = pos.long().sum(1, keepdim=True)
        num_neg = torch.clamp(self.negpos_ratio*num_pos, max=pos.size(1)-1)
        neg = idx_rank < num_neg.expand_as(idx_rank)

        # Confidence Loss Including Positive and Negative Examples
        pos_idx = pos.unsqueeze(2).expand_as(conf_data)
        neg_idx = neg.unsqueeze(2).expand_as(conf_data)
        conf_p = conf_data[(pos_idx+neg_idx).gt(0)].view(-1,self.num_classes)
        targets_weighted = conf_t[(pos+neg).gt(0)]
        loss_c = F.cross_entropy(conf_p, targets_weighted, reduction='sum')

        # Sum of losses: L(x,c,l,g) = (Lconf(x, c) + αLloc(x,l,g)) / N
        N = max(num_pos.data.sum().float(), 1)
        loss_l /= N
        loss_c /= N
        loss_landm /= N1
        return loss_l, loss_c, loss_landm

class ArcFace(nn.Module):
    def __init__(self, embedding_size, class_num, s=32.0, m=0.50, easy_margin=False):
        super(ArcFace, self).__init__()
        self.in_feature = embedding_size
        self.out_feature = class_num
        self.s = s
        self.m = m
        self.kernel = nn.Parameter(torch.Tensor(self.out_feature, self.in_feature))
        self.easy_margin = easy_margin
        self.cos_m = math.cos(m)
        self.sin_m = math.sin(m)
        self.th = math.cos(math.pi - m)
        self.mm = math.sin(math.pi - m) * m
        nn.init.xavier_uniform_(self.kernel)

    def forward(self, x, label):
        cosine = F.linear(F.normalize(x), F.normalize(self.kernel))
        # cos(theta + m)
        sine = torch.sqrt(1.0 - torch.pow(cosine, 2))
        phi = cosine * self.cos_m - sine * self.sin_m
        phi = torch.where((cosine - self.th) > 0, phi, cosine - self.mm) if not self.easy_margin else torch.where(cosine > 0, phi, cosine)
        one_hot = torch.zeros_like(cosine)
        one_hot.scatter_(1, label.view(-1, 1), 1)
        output = (one_hot * phi) + ((1.0 - one_hot) * cosine)
        output = output * self.s
        return output

class CurricularFace(nn.Module):
    def __init__(self, embedding_size, class_num, world_size=1, s=64.0, m=0.50):
        super(CurricularFace, self).__init__()
        self.in_features = embedding_size
        self.out_features = class_num
        self.m = m
        self.s = s
        self.cos_m = math.cos(m)
        self.sin_m = math.sin(m)
        self.threshold = math.cos(math.pi - m)
        self.mm = math.sin(math.pi - m) * m
        self.world_size = world_size
        self.kernel = nn.Parameter(torch.Tensor(self.in_features, self.out_features))
        self.register_buffer('t', torch.zeros(1))
        nn.init.normal_(self.kernel, std=0.01)

    def forward(self, x, label):
        kernel_norm = F.normalize(self.kernel)
        cos_theta = torch.mm(x, kernel_norm).clamp(-1, 1)
        target_logit = cos_theta[torch.arange(0, x.size(0)), label].view(-1, 1)
        sin_theta = torch.sqrt(1.0 - torch.pow(target_logit, 2))
        cos_theta_m = target_logit * self.cos_m - sin_theta * self.sin_m
        mask = cos_theta > cos_theta_m
        final_target_logit = torch.where(target_logit > self.threshold, cos_theta_m, target_logit - self.mm)
        hard_example = cos_theta[mask]
        with torch.no_grad():
            self.t = target_logit.mean() * 0.01 + (1 - 0.01) * self.t
            if self.world_size != 1:
                dist.all_reduce(self.t, dist.ReduceOp.SUM)
            self.t = self.t / self.world_size
        cos_theta[mask] = hard_example * (self.t + hard_example)
        cos_theta.scatter_(1, label.view(-1, 1).long(), final_target_logit)
        output = cos_theta * self.s
        return output


import cv2

class LineState(object):
    tableline_x_offsets = range(1, 5)
    tableline_y_offsets = range(1, 5)
    tableline_thickness = [2, 3]

    # 0/1/2/3: 仅单边（左上右下）
    # 4/5/6/7: 两边都有线（左上，右上，右下，左下）
    tableline_options = range(0, 8)

    middleline_thickness = [1, 2, 3]
    middleline_thickness_p = [0.2, 0.7, 0.1]

class Liner(object):
    def __init__(self):
        self.linestate = LineState()
        self.cfg = {}
        self.cfg['line'] = {}

        self.cfg['line']['under_line'] = {}
        self.cfg['line']['under_line']['enable'] = True
        self.cfg['line']['under_line']['fraction'] = 0.5

        self.cfg['line']['table_line'] = {}
        self.cfg['line']['table_line']['enable'] = True
        self.cfg['line']['table_line']['fraction'] = 0.5

        self.cfg['line']['middle_line'] = {}
        self.cfg['line']['middle_line']['enable'] = True
        self.cfg['line']['middle_line']['fraction'] = 0


        self.cfg['line_color'] = {}
        self.cfg['line_color']['enable'] = True

        self.cfg['line_color']['black'] = {}
        self.cfg['line_color']['black']['fraction'] = 0.5
        self.cfg['line_color']['black']['l_boundary'] = [0,0,0]
        self.cfg['line_color']['black']['h_boundary'] = [64,64,64]

        self.cfg['line_color']['blue'] = {}
        self.cfg['line_color']['blue']['fraction'] = 0.5
        self.cfg['line_color']['blue']['l_boundary'] = [0,0,150]
        self.cfg['line_color']['blue']['h_boundary'] = [60,64,255]


    def get_line_color(self):
        p = []
        colors = []
        for k, v in self.cfg['line_color'].items():
            if k == 'enable':
                continue
            p.append(v['fraction'])
            colors.append(k)

        # pick color by fraction
        color_name = np.random.choice(colors, p=p)
        l_boundary = self.cfg['line_color'][color_name]['l_boundary']
        h_boundary = self.cfg['line_color'][color_name]['h_boundary']
        # random color by low and high RGB boundary
        r = np.random.randint(l_boundary[0], h_boundary[0])
        g = np.random.randint(l_boundary[1], h_boundary[1])
        b = np.random.randint(l_boundary[2], h_boundary[2])
        return b, g, r

    def apply(self, word_img, text_box_pnts):
        """
        :param word_img:  word image with big background
        :param text_box_pnts: left-top, right-top, right-bottom, left-bottom of text word
        :return:
        """
        line_p = []
        funcs = []

        if self.cfg['line']['under_line']['enable']:
            line_p.append(self.cfg['line']['under_line']['fraction'])
            funcs.append(self.apply_under_line)

        if self.cfg['line']['table_line']['enable']:
            line_p.append(self.cfg['line']['table_line']['fraction'])
            funcs.append(self.apply_table_line)

        if self.cfg['line']['middle_line']['enable']:
            line_p.append(self.cfg['line']['middle_line']['fraction'])
            funcs.append(self.apply_middle_line)


        if len(line_p) == 0:
            return word_img, text_box_pnts

        line_effect_func = np.random.choice(funcs, p=line_p)

        if self.cfg['line_color']['enable'] or self.cfg['font_color']['enable']:
            line_color = self.get_line_color()
        else:
            line_color = word_color + random.randint(0, 10)

        return line_effect_func(word_img, text_box_pnts, line_color)

    def apply_under_line(self, word_img, text_box_pnts, line_color):
        y_offset = random.choice([0, 1])

        text_box_pnts[2][1] += y_offset
        text_box_pnts[3][1] += y_offset

        dst = cv2.line(word_img,
                       (text_box_pnts[2][0], text_box_pnts[2][1]),
                       (text_box_pnts[3][0], text_box_pnts[3][1]),
                       color=line_color,
                       thickness=3,
                       lineType=cv2.LINE_AA)

        return dst, text_box_pnts

    def apply_table_line(self, word_img, text_box_pnts, line_color):
        """
        共有 8 种可能的画法，横线横穿整张 word_img
        0/1/2/3: 仅单边（左上右下）
        4/5/6/7: 两边都有线（左上，右上，右下，左下）
        """
        dst = word_img
        option = random.choice(self.linestate.tableline_options)
        thickness = random.choice(self.linestate.tableline_thickness)

        top_y_offset = random.choice(self.linestate.tableline_y_offsets)
        bottom_y_offset = random.choice(self.linestate.tableline_y_offsets)
        left_x_offset = random.choice(self.linestate.tableline_x_offsets)
        right_x_offset = random.choice(self.linestate.tableline_x_offsets)

        def is_top():
            return option in [1, 4, 5]

        def is_bottom():
            return option in [3, 6, 7]

        def is_left():
            return option in [0, 4, 7]

        def is_right():
            return option in [2, 5, 6]

        if is_top():
            text_box_pnts[0][1] -= top_y_offset
            text_box_pnts[1][1] -= top_y_offset

        if is_bottom():
            text_box_pnts[2][1] += bottom_y_offset
            text_box_pnts[3][1] += bottom_y_offset

        if is_left():
            text_box_pnts[0][0] -= left_x_offset
            text_box_pnts[3][0] -= left_x_offset

        if is_right():
            text_box_pnts[1][0] += right_x_offset
            text_box_pnts[2][0] += right_x_offset

        if is_bottom():
            dst = cv2.line(dst,
                           (0, text_box_pnts[2][1]),
                           (word_img.shape[1], text_box_pnts[3][1]),
                           color=line_color,
                           thickness=thickness,
                           lineType=cv2.LINE_AA)

        if is_top():
            dst = cv2.line(dst,
                           (0, text_box_pnts[0][1]),
                           (word_img.shape[1], text_box_pnts[1][1]),
                           color=line_color,
                           thickness=thickness,
                           lineType=cv2.LINE_AA)

        if is_left():
            dst = cv2.line(dst,
                           (text_box_pnts[0][0], 0),
                           (text_box_pnts[3][0], word_img.shape[0]),
                           color=line_color,
                           thickness=thickness,
                           lineType=cv2.LINE_AA)

        if is_right():
            dst = cv2.line(dst,
                           (text_box_pnts[1][0], 0),
                           (text_box_pnts[2][0], word_img.shape[0]),
                           color=line_color,
                           thickness=thickness,
                           lineType=cv2.LINE_AA)

        return dst, text_box_pnts

    def apply_middle_line(self, word_img, text_box_pnts, line_color):
        y_center = int((text_box_pnts[0][1] + text_box_pnts[3][1]) / 2)

        thickness = np.random.choice(self.linestate.middleline_thickness, p=self.linestate.middleline_thickness_p)

        dst = cv2.line(word_img,
                       (text_box_pnts[0][0], y_center),
                       (text_box_pnts[1][0], y_center),
                       color=line_color,
                       thickness=thickness,
                       lineType=cv2.LINE_AA)

        return dst, text_box_pnts


# -*- coding:utf-8 -*-
import math
import torch
import torch.nn as nn
import torch.nn.functional as F

import torch.distributed as dist

class LossBase(object):
    def __init__(self, deepvac_config):
        self.auditConfig()

    def auditConfig(self):
        raise Exception("Not implemented!")

    def __call__(self,img):
        raise Exception("Not implemented!")


class MaskL1Loss(LossBase):
    def __init__(self, deepvac_config):
        super(MaskL1Loss, self).__init__(deepvac_config)

    def auditConfig(self):
        self.eps = 1e-6

    def __call__(self, pred: torch.Tensor, gt, mask):
        loss = (torch.abs(pred - gt) * mask).sum() / (mask.sum() + self.eps)
        return loss

class DiceLoss(LossBase):
    def __init__(self, deepvac_config):
        super(DiceLoss, self).__init__(deepvac_config)

    def auditConfig(self):
        self.eps = 1e-6

    def __call__(self, pred: torch.Tensor, gt, mask, weights=None):
        return self._compute(pred, gt, mask, weights)

    def _compute(self, pred, gt, mask, weights):
        if pred.dim() == 4:
            pred = pred[:, 0, :, :]
            gt = gt[:, 0, :, :]
        assert pred.shape == gt.shape
        assert pred.shape == mask.shape
        if weights is not None:
            assert weights.shape == mask.shape
            mask = weights * mask
        intersection = (pred * gt * mask).sum()
        union = (pred * mask).sum() + (gt * mask).sum() + self.eps
        loss = 1 - 2.0 * intersection / union
        assert loss <= 1
        return loss

class BalanceCrossEntropyLoss(LossBase):
    def __init__(self, deepvac_config):
        super(BalanceCrossEntropyLoss, self).__init__(deepvac_config)

    def auditConfig(self):
        self.negative_ratio = 3.0
        self.eps = 1e-6

    def __call__(self,
                pred: torch.Tensor,
                gt: torch.Tensor,
                mask: torch.Tensor,
                return_origin=False):
        positive = (gt * mask).byte()
        negative = ((1 - gt) * mask).byte()
        positive_count = int(positive.float().sum())
        negative_count = min(int(negative.float().sum()), int(positive_count * self.negative_ratio))
        loss = nn.functional.binary_cross_entropy(pred, gt, reduction='none')
        positive_loss = loss * positive.float()
        negative_loss = loss * negative.float()
        negative_loss, _ = negative_loss.view(-1).topk(negative_count)
        balance_loss = (positive_loss.sum() + negative_loss.sum()) / (positive_count + negative_count + self.eps)
        if return_origin:
            return balance_loss, loss
        return balance_loss

class BCEBlurWithLogitsLoss(LossBase):
    # BCEwithLogitLoss() with reduced missing label effects.
    def __init__(self, deepvac_config):
        super(BCEBlurWithLogitsLoss, self).__init__(deepvac_config)
        self.loss_fcn = nn.BCEWithLogitsLoss(reduction=self.reduction)

    def auditConfig(self):
        self.alpha = 0.05
        self.reduction = 'none'

    def __call__(self, pred, true):
        loss = self.loss_fcn(pred, true)
        pred = torch.sigmoid(pred)
        # reduce only missing label effects
        dx = pred - true
        alpha_factor = 1 - torch.exp((dx - 1) / (self.alpha + 1e-4))
        loss *= alpha_factor
        return loss.mean()

class FocalLoss(LossBase):
    # Wraps focal loss around existing loss_fcn(), i.e. criteria = FocalLoss(nn.BCEWithLogitsLoss(), gamma=1.5)
    def __init__(self, deepvac_config):
        super(FocalLoss, self).__init__(deepvac_config)
        # must be nn.BCEWithLogitsLoss()
        self.reduction = deepvac_config.reduction
        self.loss_fcn = nn.BCEWithLogitsLoss(reduction='none')

    def auditConfig(self):
        self.gamma = 1.5
        self.alpha = 0.25

    def __call__(self, pred, true):
        loss = self.loss_fcn(pred, true)
        # TF implementation https://github.com/tensorflow/addons/blob/v0.7.1/tensorflow_addons/losses/focal_loss.py
        pred_prob = torch.sigmoid(pred)
        p_t = true * pred_prob + (1 - true) * (1 - pred_prob)
        alpha_factor = true * self.alpha + (1 - true) * (1 - self.alpha)
        modulating_factor = (1.0 - p_t) ** self.gamma
        loss *= alpha_factor * modulating_factor

        if self.reduction == 'mean':
            return loss.mean()
        elif self.reduction == 'sum':
            return loss.sum()
        else:
            return loss

class QFocalLoss(LossBase):
    # Wraps Quality focal loss around existing loss_fcn(), i.e. criteria = FocalLoss(nn.BCEWithLogitsLoss(), gamma=1.5)
    def __init__(self, deepvac_config):
        super(QFocalLoss, self).__init__(deepvac_config)
        # must be nn.BCEWithLogitsLoss()
        self.loss_fcn = deepvac_config.loss_fcn

    def auditConfig(self):
        self.gamma = 1.5
        self.alpha = 0.25
        self.reduction = 'none'

    def __call__(self, pred, true):
        loss = self.loss_fcn(pred, true)
        pred_prob = torch.sigmoid(pred)
        alpha_factor = true * self.alpha + (1 - true) * (1 - self.alpha)
        modulating_factor = torch.abs(true - pred_prob) ** self.gamma
        loss *= alpha_factor * modulating_factor

        if self.reduction == 'mean':
            return loss.mean()
        elif self.reduction == 'sum':
            return loss.sum()
        else:
            return loss

class WingLoss(nn.Module):
    def __init__(self):
        super(WingLoss, self).__init__()

    def forward(self, pred, truth, w=10.0, epsilon=2.0):
        x = truth - pred
        c = w * (1.0 - math.log(1.0 + w / epsilon))
        absolute_x = torch.abs(x)
        losses = torch.where(w > absolute_x, w * torch.log(1.0 + absolute_x / epsilon), absolute_x - c)
        return torch.sum(losses) / (len(losses) * 1.0)



#function only be executed once.
def syszux_once(f):
    def wrapper(*args, **kwargs):
        if not wrapper.has_run:
            wrapper.has_run = True
            return f(*args, **kwargs)
    wrapper.has_run = False
    return wrapper

if __name__ == "__main__":
    @syszux_once
    def test(x,y):
        print(x**y)

    test(2,3)
    test(3,4)
    


from ..utils import LOG
from .base import DeepvacCast

class CoremlCast(DeepvacCast):
    def __init__(self, deepvac_train_config):
        super(CoremlCast,self).__init__(deepvac_train_config)

    def auditConfig(self):
        if not self.config.coreml_model_dir:
            return False

        if not self.config.trace_model_dir and not self.config.script_model_dir:
            LOG.logE("CoreML converter now has dependency on TorchScript model, you need to enable config.train.trace_model_dir or config.train.script_model_dir", exit=True)

        if self.config.coreml_input_type not in [None, 'image','tensor']:
            LOG.logE("coreml input type must be {}".format([None, 'image','tensor']), exit=True)

        if self.config.coreml_input_type == 'image':
            LOG.logI("You are in coreml_input_type=image mod")
            if self.config.coreml_scale is None:
                LOG.logE("You need to set config.train.coreml_scale in config.py, e.g. config.train.coreml_scale = 1.0 / (0.226 * 255.0)", exit=True)
            
            if self.config.coreml_color_layout is None:
                LOG.logE("You need to set config.train.coreml_color_layout in config.py, e.g. config.train.coreml_color_layout = 'BGR' ", exit=True)

            if self.config.coreml_blue_bias is None:
                LOG.logE("You need to set config.train.coreml_blue_bias in config.py, e.g. config.train.coreml_blue_bias = -0.406 / 0.226 ", exit=True)

            if self.config.coreml_green_bias is None:
                LOG.logE("You need to set config.train.coreml_green_bias in config.py, e.g. config.train.coreml_green_bias = -0.456 / 0.226", exit=True)

            if self.config.coreml_red_bias is None:
                LOG.logE("You need to set config.train.coreml_red_bias in config.py, e.g. config.train.coreml_red_bias = -0.485 / 0.226 ", exit=True)

        return True

    def process(self, cast_output_file=None):
        try:
            import coremltools
        except:
            LOG.logE("You need to install coremltools package if you want to convert PyTorch to CoreML model. E.g. pip install --upgrade coremltools")
            return

        output_coreml_file = self.config.coreml_model_dir
        if cast_output_file:
            output_coreml_file = '{}/coreml__{}.mlmodel'.format(self.config.output_dir, cast_output_file)
            self.config.coreml_model_dir = output_coreml_file

        LOG.logI("config.coreml_model_dir found, save coreml model to {}...".format(self.config.coreml_model_dir))
        model = self.config.script_model_dir
        if self.config.script_model_dir is None:
            model = self.config.trace_model_dir
        #input mode
        if self.config.coreml_input_type == 'image':
            input = coremltools.ImageType(name="input",
                        shape=tuple(self.config.sample.shape),
                        scale=self.config.coreml_scale,
                        color_layout = self.config.coreml_color_layout,
                        bias = [self.config.coreml_blue_bias, self.config.coreml_green_bias, self.config.coreml_red_bias])
        else:
            input = coremltools.TensorType(name='input', shape=tuple(self.config.sample.shape))
        #convert
        coreml_model = coremltools.convert(model=model, inputs=[input], 
             classfier_config=self.config.coreml_classfier_config, minimum_deployment_target=self.config.coreml_minimum_deployment_target)

        # Set feature descriptions (these show up as comments in XCode)
        coreml_model.input_description["input"] = "Deepvac Model Input"
        coreml_model.output_description["classLabel"] = "Most likely image category"

        # Set model author name
        coreml_model.author = '"DeepVAC'

        # Set the license of the model
        coreml_model.license = "Deepvac Lincense"
        coreml_model.short_description = "Powered by DeepVAC"

        # Set a version for the model
        coreml_model.version = self.config.coreml_version if self.config.coreml_version else "1.0"
        # Save the CoreML model
        coreml_model.save(output_coreml_file)


import cv2
import numpy as np
import random
from .base_aug import AugBase

# 随机旋转（针对于人脸关键点任务）
class RandomRotateFacialKpListAug(AugBase):
    def __init__(self, deepvac_config):
        super(RandomRotateFacialKpListAug, self).__init__(deepvac_config)

    def auditConfig(self):
        pass

    # img: include 2 element
    # img[0] -> img
    # img[1] -> keypoint info (scale to 0-1)
    def __call__(self, img):
        img,landmarks = self.auditInput(img, has_label=True)
        angle = random.choice([-13, -12, -11, -10, -9, -8, -7, -6, -5, 0, 2, 1, 2, 3, 4, 5, 6, 7, 8, 9, 11, 10, 11, 12, 13])
        h, w, _ = img.shape
        M = cv2.getRotationMatrix2D((w / 2, h / 2), angle, 1)
        dest_img = cv2.warpAffine(img, M, (w, h))
        dest_landmarks = []
        for i in range(len(landmarks)):
            curlandmark_x = landmarks[i][0] * w
            curlandmark_y = landmarks[i][1] * h
            dst_x = curlandmark_x * M[0][0] + curlandmark_y * M[0][1] + M[0][2]
            dst_y = curlandmark_x * M[1][0] + curlandmark_y * M[1][1] + M[1][2]
            dest_landmarks.append([dst_x / w, dst_y / h])
        return [dest_img, dest_landmarks]

# 随机水平翻转（针对于人脸关键点任务,关键点索引顺序等需要和deepvac人脸关键点索引顺序一致）
class RandomFilpFacialKpListAug(AugBase):
    def __init__(self, deepvac_config):
        super(RandomFilpFacialKpListAug, self).__init__(deepvac_config)

    def auditConfig(self):
        pass

    def flipLandmark(self, dest_landmark, src_landmark, sequences):
        for sequence in sequences:
            for i in range(sequence[1], sequence[0] - 1, -1):
                dest_landmark.append(src_landmark[i])
        return dest_landmark

    def copyLandmark(self, dest_landmark, src_landmark, sequences):
        for sequence in sequences:
            for i in range(sequence[0], sequence[1] + 1):
                dest_landmark.append(src_landmark[i])
        return dest_landmark

    # img: include 2 element
    # img[0] -> img
    # img[1] -> keypoint info(scale to 0-1)
    def __call__(self, img):
        img,landmarks = self.auditInput(img, has_label=True)
        h, w, _ = img.shape
        dest_img = cv2.flip(img, 1)
        flip_landmarks = []
        for i in range(len(landmarks)):
            curlandmark_x = w - 1 - landmarks[i][0]
            curlandmark_y = landmarks[i][1]
            flip_landmarks.append([curlandmark_x, curlandmark_y])

        dest_landmarks = []
        # 重新排列关键点顺序
        flip_landmarks_list = [[0, 32], [38, 42], [33, 37]]
        dest_landmarks = self.flipLandmark(dest_landmarks, flip_landmarks, flip_landmarks_list)
        dest_landmarks = self.copyLandmark(dest_landmarks, flip_landmarks, [[43, 46]])
        flip_landmarks_list = [[47, 51], [58, 61], [62, 63], [52, 55], [56, 57], [68, 71], [64, 67]]
        dest_landmarks = self.flipLandmark(dest_landmarks, flip_landmarks, flip_landmarks_list)
        dest_landmarks = self.copyLandmark(dest_landmarks, flip_landmarks, [[75, 77], [72, 74]])
        flip_landmarks_list = [[78, 79], [80, 81], [82, 83], [84, 90], [91, 95], [96, 100], [101, 103], [104, 105]]
        dest_landmarks = self.flipLandmark(dest_landmarks, flip_landmarks, flip_landmarks_list)

        return [dest_img, dest_landmarks]

class CropFacialWithBoxesAndLmksAug(AugBase):
    def __init__(self, deepvac_config):
        super(CropFacialWithBoxesAndLmksAug, self).__init__(deepvac_config)

    def auditConfig(self):
        self.config.facial_img_dim = self.addUserConfig('facial_img_dim', self.config.facial_img_dim, 640, True)

    def _matrix_iof(self, a, b):
        lt = np.maximum(a[:, np.newaxis, :2], b[:, :2])
        rb = np.minimum(a[:, np.newaxis, 2:], b[:, 2:])
        area_i = np.prod(rb - lt, axis=2) * (lt < rb).all(axis=2)
        area_a = np.prod(a[:, 2:] - a[:, :2], axis=1)
        return area_i / np.maximum(area_a[:, np.newaxis], 1)

    def __call__(self, image):
        image, label = self.auditInput(image, has_label=True)
        assert isinstance(label, list) and len(label) == 3, "label must be list, and length should be 3"
        assert isinstance(label[0], np.ndarray) and label[0].ndim == 2, "label[0](boxes) must be numpy.ndarray, and shape should be (n, 4)"
        assert isinstance(label[1], np.ndarray) and label[1].ndim == 2, "label[1](landms) must be numpy.ndarray, and shape should be (n, 10)"
        assert isinstance(label[2], np.ndarray) and label[2].ndim == 1, "label[2](labels) must be numpy.ndarray, and shape should be (n, )"

        boxes, landms, labels = label
        height, width, _ = image.shape

        for _ in range(250):

            PRE_SCALES = [0.3, 0.45, 0.6, 0.8, 1.0]
            scale = random.choice(PRE_SCALES)
            short_side = min(width, height)
            w = int(scale * short_side)
            h = w

            if width == w:
                l = 0
            else:
                l = random.randrange(width - w)
            if height == h:
                t = 0
            else:
                t = random.randrange(height - h)
            roi = np.array((l, t, l + w, t + h))

            value = self._matrix_iof(boxes, roi[np.newaxis])
            flag = (value >= 1)
            if not flag.any():
                continue

            centers = (boxes[:, :2] + boxes[:, 2:]) / 2
            mask_a = np.logical_and(roi[:2] < centers, centers < roi[2:]).all(axis=1)
            boxes_t = boxes[mask_a].copy()
            labels_t = labels[mask_a].copy()
            landms_t = landms[mask_a].copy()
            landms_t = landms_t.reshape([-1, 5, 2])

            if boxes_t.shape[0] == 0:
                continue

            image_t = image[roi[1]:roi[3], roi[0]:roi[2]]

            boxes_t[:, :2] = np.maximum(boxes_t[:, :2], roi[:2])
            boxes_t[:, :2] -= roi[:2]
            boxes_t[:, 2:] = np.minimum(boxes_t[:, 2:], roi[2:])
            boxes_t[:, 2:] -= roi[:2]

            # landm
            landms_t[:, :, :2] = landms_t[:, :, :2] - roi[:2]
            landms_t[:, :, :2] = np.maximum(landms_t[:, :, :2], np.array([0, 0]))
            landms_t[:, :, :2] = np.minimum(landms_t[:, :, :2], roi[2:] - roi[:2])
            landms_t = landms_t.reshape([-1, 10])


            # make sure that the cropped image contains at least one face > 16 pixel at training image scale
            b_w_t = (boxes_t[:, 2] - boxes_t[:, 0] + 1) / w * self.config.facial_img_dim
            b_h_t = (boxes_t[:, 3] - boxes_t[:, 1] + 1) / h * self.config.facial_img_dim
            mask_b = np.minimum(b_w_t, b_h_t) > 0.0
            boxes_t = boxes_t[mask_b]
            labels_t = labels_t[mask_b]
            landms_t = landms_t[mask_b]

            if boxes_t.shape[0] == 0:
                continue

            return image_t, [boxes_t, landms_t, labels_t]
        return image, [boxes, landms, labels]


class DistortFacialAugBase(AugBase):
    def __init__(self, deepvac_config):
        super(DistortFacialAugBase, self).__init__(deepvac_config)

    def _convert(self, image, alpha=1, beta=0):
        tmp = image.astype(float) * alpha + beta
        tmp[tmp < 0] = 0
        tmp[tmp > 255] = 255
        image[:] = tmp

    def auditConfig(self):
        pass

class BrightDistortFacialAug(DistortFacialAugBase):
    def __init__(self, deepvac_config):
        super(BrightDistortFacialAug, self).__init__(deepvac_config)

    def auditConfig(self):
        pass

    def __call__(self, image):
        image, label = self.auditInput(image, has_label=True)
        assert isinstance(label, list) and len(label) == 3, "label must be list, and length should be 3"
        assert isinstance(label[0], np.ndarray) and label[0].ndim == 2, "label[0](boxes) must be numpy.ndarray, and shape should be (n, 4)"
        assert isinstance(label[1], np.ndarray) and label[1].ndim == 2, "label[1](landms) must be numpy.ndarray, and shape should be (n, 10)"
        assert isinstance(label[2], np.ndarray) and label[2].ndim == 1, "label[2](labels) must be numpy.ndarray, and shape should be (n, )"

        self._convert(image, beta=random.uniform(-32, 32))

        return image, label

class ContrastDistortFacialAug(DistortFacialAugBase):
    def __init__(self, deepvac_config):
        super(ContrastDistortFacialAug, self).__init__(deepvac_config)

    def auditConfig(self):
        pass

    def __call__(self, image):
        image, label = self.auditInput(image, has_label=True)
        assert isinstance(label, list) and len(label) == 3, "label must be list, and length should be 3"
        assert isinstance(label[0], np.ndarray) and label[0].ndim == 2, "label[0](boxes) must be numpy.ndarray, and shape should be (n, 4)"
        assert isinstance(label[1], np.ndarray) and label[1].ndim == 2, "label[1](landms) must be numpy.ndarray, and shape should be (n, 10)"
        assert isinstance(label[2], np.ndarray) and label[2].ndim == 1, "label[2](labels) must be numpy.ndarray, and shape should be (n, )"

        self._convert(image, alpha=random.uniform(0.5, 1.5))

        return image, label

class SaturationDistortFacialAug(DistortFacialAugBase):
    def __init__(self, deepvac_config):
        super(SaturationDistortFacialAug, self).__init__(deepvac_config)

    def auditConfig(self):
        pass

    def __call__(self, image):
        image, label = self.auditInput(image, has_label=True)
        assert isinstance(label, list) and len(label) == 3, "label must be list, and length should be 3"
        assert isinstance(label[0], np.ndarray) and label[0].ndim == 2, "label[0](boxes) must be numpy.ndarray, and shape should be (n, 4)"
        assert isinstance(label[1], np.ndarray) and label[1].ndim == 2, "label[1](landms) must be numpy.ndarray, and shape should be (n, 10)"
        assert isinstance(label[2], np.ndarray) and label[2].ndim == 1, "label[2](labels) must be numpy.ndarray, and shape should be (n, )"

        image = cv2.cvtColor(image, cv2.COLOR_BGR2HSV)
        self._convert(image[:, :, 1], alpha=random.uniform(0.5, 1.5))
        image = cv2.cvtColor(image, cv2.COLOR_HSV2BGR)

        return image, label

class HueDistortFacialAug(DistortFacialAugBase):
    def __init__(self, deepvac_config):
        super(HueDistortFacialAug, self).__init__(deepvac_config)

    def auditConfig(self):
        pass

    def __call__(self, image):
        image, label = self.auditInput(image, has_label=True)
        assert isinstance(label, list) and len(label) == 3, "label must be list, and length should be 3"
        assert isinstance(label[0], np.ndarray) and label[0].ndim == 2, "label[0](boxes) must be numpy.ndarray, and shape should be (n, 4)"
        assert isinstance(label[1], np.ndarray) and label[1].ndim == 2, "label[1](landms) must be numpy.ndarray, and shape should be (n, 10)"
        assert isinstance(label[2], np.ndarray) and label[2].ndim == 1, "label[2](labels) must be numpy.ndarray, and shape should be (n, )"

        image = cv2.cvtColor(image, cv2.COLOR_BGR2HSV)
        tmp = image[:, :, 0].astype(int) + random.randint(-18, 18)
        tmp %= 180
        image[:, :, 0] = tmp
        image = cv2.cvtColor(image, cv2.COLOR_HSV2BGR)

        return image, label

class MirrorFacialAug(AugBase):
    def __init__(self, deepvac_config):
        super(MirrorFacialAug, self).__init__(deepvac_config)

    def auditConfig(self):
        pass

    def __call__(self, image):
        image, label = self.auditInput(image, has_label=True)
        assert isinstance(label, list) and len(label) == 3, "label must be list, and length should be 3"
        assert isinstance(label[0], np.ndarray) and label[0].ndim == 2, "label[0](boxes) must be numpy.ndarray, and shape should be (n, 4)"
        assert isinstance(label[1], np.ndarray) and label[1].ndim == 2, "label[1](landms) must be numpy.ndarray, and shape should be (n, 10)"
        assert isinstance(label[2], np.ndarray) and label[2].ndim == 1, "label[2](labels) must be numpy.ndarray, and shape should be (n, )"

        boxes, landms, labels = label
        _, width, _ = image.shape
        image = image[:, ::-1]
        boxes = boxes.copy()
        boxes[:, 0::2] = width - boxes[:, 2::-2]

        # landm
        landms = landms.copy()
        landms = landms.reshape([-1, 5, 2])
        landms[:, :, 0] = width - landms[:, :, 0]
        tmp = landms[:, 1, :].copy()
        landms[:, 1, :] = landms[:, 0, :]
        landms[:, 0, :] = tmp
        tmp1 = landms[:, 4, :].copy()
        landms[:, 4, :] = landms[:, 3, :]
        landms[:, 3, :] = tmp1
        landms = landms.reshape([-1, 10])

        return image, [boxes, landms, labels]

class Pad2SquareFacialAug(AugBase):
    def __init__(self, deepvac_config):
        super(Pad2SquareFacialAug, self).__init__(deepvac_config)

    def auditConfig(self):
        self.config.facial_rgb_means = self.addUserConfig('facial_rgb_means', self.config.facial_rgb_means, (104, 117, 123))

    def __call__(self, image):
        image, label = self.auditInput(image, has_label=True)
        assert isinstance(label, list) and len(label) == 3, "label must be list, and length should be 3"
        assert isinstance(label[0], np.ndarray) and label[0].ndim == 2, "label[0](boxes) must be numpy.ndarray, and shape should be (n, 4)"
        assert isinstance(label[1], np.ndarray) and label[1].ndim == 2, "label[1](landms) must be numpy.ndarray, and shape should be (n, 10)"
        assert isinstance(label[2], np.ndarray) and label[2].ndim == 1, "label[2](labels) must be numpy.ndarray, and shape should be (n, )"

        height, width, _ = image.shape
        if height == width:
            return image, label
        long_side = max(width, height)
        image_t = np.empty((long_side, long_side, 3), dtype=image.dtype)
        image_t[:, :] = self.config.facial_rgb_means
        image_t[0:0 + height, 0:0 + width] = image
        return image, label

class ResizeSubtractMeanFacialAug(AugBase):
    def __init__(self, deepvac_config):
        super(ResizeSubtractMeanFacialAug, self).__init__(deepvac_config)

    def auditConfig(self):
        self.config.facial_img_dim = self.addUserConfig('facial_img_dim', self.config.facial_img_dim, 640, True)
        self.config.facial_rgb_means = self.addUserConfig('facial_rgb_means', self.config.facial_rgb_means, (104, 117, 123))

    def __call__(self, image):
        image, label = self.auditInput(image, has_label=True)
        assert isinstance(label, list) and len(label) == 3, "label must be list, and length should be 3"
        assert isinstance(label[0], np.ndarray) and label[0].ndim == 2, "label[0](boxes) must be numpy.ndarray, and shape should be (n, 4)"
        assert isinstance(label[1], np.ndarray) and label[1].ndim == 2, "label[1](landms) must be numpy.ndarray, and shape should be (n, 10)"
        assert isinstance(label[2], np.ndarray) and label[2].ndim == 1, "label[2](labels) must be numpy.ndarray, and shape should be (n, )"

        boxes, landms, labels = label
        height, width, _ = image.shape

        interp_methods = [cv2.INTER_LINEAR, cv2.INTER_CUBIC, cv2.INTER_AREA, cv2.INTER_NEAREST, cv2.INTER_LANCZOS4]
        interp_method = interp_methods[random.randrange(5)]
        image = cv2.resize(image, (self.config.facial_img_dim, self.config.facial_img_dim), interpolation=interp_method)
        image = image.astype(np.float32)
        image -= self.config.facial_rgb_means

        boxes[:, 0::2] /= width
        boxes[:, 1::2] /= height

        landms[:, 0::2] /= width
        landms[:, 1::2] /= height

        return image.transpose(2, 0, 1), [boxes, landms, labels]